import typing
from collections import deque

from kungfu import Error, Ok, Pulse, Result

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.dispatch.handler.func import FuncHandler, Function
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.process import check_rule, process_inner
from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager
from telegrinder.bot.dispatch.view.abc import ABCView
from telegrinder.bot.rules.abc import ABCRule, Always
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import (
    BusinessConnection,
    BusinessMessagesDeleted,
    CallbackQuery,
    ChatBoostRemoved,
    ChatBoostUpdated,
    ChatJoinRequest,
    ChatMemberUpdated,
    ChosenInlineResult,
    InlineQuery,
    Message,
    MessageReactionCountUpdated,
    MessageReactionUpdated,
    PaidMediaPurchased,
    Poll,
    PollAnswer,
    PreCheckoutQuery,
    ShippingQuery,
    Update,
)

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

type UpdateModel = typing.Union[
    BusinessConnection,
    BusinessMessagesDeleted,
    CallbackQuery,
    ChatBoostRemoved,
    ChatBoostUpdated,
    ChatJoinRequest,
    ChatMemberUpdated,
    ChosenInlineResult,
    InlineQuery,
    Message,
    MessageReactionCountUpdated,
    MessageReactionUpdated,
    PaidMediaPurchased,
    Poll,
    PollAnswer,
    PreCheckoutQuery,
    ShippingQuery,
]
type ViewResult = Result[str, str] | Result[str, Exception]

OK_CHECK: typing.Final = Ok()


class View(ABCView):
    filter: ABCRule
    handlers: deque[ABCHandler]
    middlewares: deque[ABCMiddleware]
    return_manager: ABCReturnManager | None

    def __init__(
        self,
        *,
        agent_cls: type[Agent] | None = None,
        return_manager: ABCReturnManager | None = None,
    ) -> None:
        self.filter = Always()
        self.handlers = deque()
        self.middlewares = deque()
        self.agent_cls = agent_cls
        self.return_manager = return_manager

    def __bool__(self) -> bool:
        return bool(self.handlers) or bool(self.middlewares)

    def __repr__(self) -> str:
        return "<{}>".format(type(self).__name__)

    @property
    def auto_rules(self) -> ABCRule:
        return self.filter

    @auto_rules.setter
    def auto_rules(self, value: ABCRule | typing.Iterable[ABCRule], /) -> None:
        for rule in (value,) if isinstance(value, ABCRule) else value:
            self.filter &= rule

    def __call__[T: Function](
        self,
        *rules: ABCRule,
        final: bool = True,
        agent: type[Agent] | None = None,
    ) -> typing.Callable[[T], T]:
        def decorator(function: T, /) -> T:
            self.handlers.append(
                FuncHandler(
                    function=function,
                    rules=rules,
                    agent=agent or self.agent_cls,
                    final=final,
                ),
            )
            return function

        return decorator

    @typing.overload
    def register_middleware[T: ABCMiddleware](self, middleware_cls: type[T], /) -> type[T]: ...

    @typing.overload
    def register_middleware(self, middleware: ABCMiddleware, /) -> None: ...

    def register_middleware(
        self,
        middleware: type[ABCMiddleware] | ABCMiddleware,
    ) -> typing.Callable[..., typing.Any] | None:
        self.middlewares.append(middleware() if isinstance(middleware, type) else middleware)
        return middleware if isinstance(middleware, type) else None

    def load(self, external: typing.Self, /) -> None:
        if not external:
            return

        self.handlers.extend(external.handlers)
        self.middlewares.extend(external.middlewares)

        if not isinstance(external.filter, Always):
            self.filter &= external.filter

        if external.return_manager is not None and self.return_manager is None:
            self.return_manager = external.return_manager

        elif self.return_manager is not None and external.return_manager is not None:
            self.return_manager.managers.extend(external.return_manager.managers)

    async def check(self, api: API, update: Update, context: Context) -> Pulse[str]:
        if not bool(self):
            return Error("View is empty.")

        if not await check_rule(self.filter, context):
            return Error("Filter is failed.")

        return OK_CHECK

    async def process(self, api: API, update: Update, context: Context) -> ViewResult:
        return await process_inner(
            api,
            update,
            context,
            self,
        )


class EventView(View):
    def __init__(
        self,
        update_type: UpdateType,
        return_manager: ABCReturnManager | None = None,
        agent_cls: type[Agent] | None = None,
    ) -> None:
        super().__init__(
            agent_cls=agent_cls,
            return_manager=return_manager,
        )
        self.update_type = update_type

    def __repr__(self) -> str:
        return "<{}: {!r}>".format(type(self).__name__, self.update_type)

    async def check(self, api: API, update: Update, context: Context) -> Pulse[str]:
        # If update is not of the expected, instantly skip checking the view
        if update.update_type != self.update_type:
            return Error(f"Incoming event `{update.update_type!r}` is not `{self.update_type!r}`.")
        return await super().check(api, update, context)


class EventModelView[T: (UpdateModel)](View):
    def __init__(
        self,
        model: type[T],
        return_manager: ABCReturnManager | None = None,
        agent_cls: type[Agent] | None = None,
    ) -> None:
        super().__init__(
            agent_cls=agent_cls,
            return_manager=return_manager,
        )
        self.model = model

    def __repr__(self) -> str:
        return "<{}: {}>".format(type(self).__name__, self.model.__name__)

    async def check(self, api: API, update: Update, context: Context) -> Pulse[str]:
        # If update object is not of the expected type of object, instantly skip checking the view
        if update.incoming_update.__class__ is not self.model:
            return Error(
                f"Incoming event model `{update.incoming_update.__class__.__name__!r}`"
                f" is not `{self.model.__name__!r}`.",
            )
        return await super().check(api, update, context)


class ErrorView(View):
    async def process(self, api: API, update: Update, context: Context) -> ViewResult:
        result = await super().process(api, update, context)

        if not result and context.exception_update:
            return Error(context.exception_update.unwrap())

        return result


class RawEventView(View):
    pass


__all__ = ("ErrorView", "EventView", "RawEventView", "View")



from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.utils import MEDIA_TYPES
from telegrinder.bot.rules.abc import ABCRule


class IsMediaGroup(ABCRule):
    def check(self, message: MessageCute) -> bool:
        if not message.media_group_id:
            return False
        return message.content_type in MEDIA_TYPES


__all__ = ("IsMediaGroup",)



from http import HTTPStatus

from kungfu import Error, Ok, Result

from telegrinder.api.api import API
from telegrinder.api.error import InvalidTokenError
from telegrinder.api.token import Token


async def validate_token(token: str, /) -> Result[Token, InvalidTokenError]:
    try:
        token = Token(token)
        api = API(token)

        match await api.get_me():
            case Ok(_):
                return Ok(token)
            case Error(error):
                return Error(
                    InvalidTokenError(
                        "Token seems to be invalid."
                        if error.status_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.NOT_FOUND)
                        else f"Unknown error {error!r} while validating token, please try again later.",
                    ),
                )
    except InvalidTokenError as error:
        return Error(error)


__all__ = ("validate_token",)



import dataclasses
import threading
import typing
from copy import deepcopy
from functools import cached_property, wraps

from kungfu.library.misc import from_optional
from kungfu.library.monad import Error, Nothing, Ok, Option, Result, Some

from telegrinder.modules import logger
from telegrinder.msgspec_utils import convert
from telegrinder.tools.fullname import fullname
from telegrinder.tools.global_context.abc import NOVALUE, ABCGlobalContext, CtxVar, CtxVariable, GlobalCtxVar

if typing.TYPE_CHECKING:
    _: typing.TypeAlias = None
else:
    _ = lambda: None


def type_check[T = typing.Any](value: typing.Any, value_type: type[T], /) -> typing.TypeGuard[T]:
    if value_type in (typing.Any, object):
        return True

    match convert(value, value_type):
        case Ok(v):
            return type(value) is type(v)
        case Error(_):
            return False

    return False


def is_dunder(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def get_orig_class[T = typing.Any](obj: T) -> type[T]:
    if "__orig_class__" not in obj.__dict__:
        return type(obj)
    return obj.__dict__["__orig_class__"]


def root_protection[F: typing.Callable[..., typing.Any]](func: F, /) -> F:
    if func.__name__ not in ("__setattr__", "__getattr__", "__delattr__"):
        raise RuntimeError(
            "You cannot decorate a {!r} function with this decorator, only "
            "'__setattr__', __getattr__', '__delattr__' methods.".format(
                func.__name__,
            )
        )

    @wraps(func)
    def wrapper(self: GlobalContext, name: str, /, *args: typing.Any) -> typing.Any:
        if self.is_root_attribute(name) and name in (self.__dict__ | self.__class__.__dict__):
            root_attr = self.get_root_attribute(name).unwrap()
            if all((not root_attr.can_be_rewritten, not root_attr.can_be_read)):
                raise AttributeError(f"Unable to set, get, delete root attribute {name!r}.")
            if func.__name__ == "__setattr__" and not root_attr.can_be_rewritten:
                raise AttributeError(f"Unable to set root attribute {name!r}.")
            if func.__name__ == "__getattr__" and not root_attr.can_be_read:
                raise AttributeError(f"Unable to get root attribute {name!r}.")
            if func.__name__ == "__delattr__":
                raise AttributeError(f"Unable to delete root attribute {name!r}.")

        return func(self, name, *args)  # type: ignore

    return wrapper  # type: ignore


@typing.overload
def ctx_var() -> typing.Any: ...


@typing.overload
def ctx_var(
    *,
    init: bool = ...,
    const: bool = ...,
) -> typing.Any: ...


@typing.overload
def ctx_var(
    *,
    default: typing.Any,
    init: bool = ...,
    const: bool = ...,
) -> typing.Any: ...


@typing.overload
def ctx_var(
    *,
    default_factory: typing.Callable[[], typing.Any],
    init: bool = ...,
    const: bool = ...,
) -> typing.Any: ...


def ctx_var(
    *,
    default: typing.Any = NOVALUE,
    default_factory: typing.Any = NOVALUE,
    const: bool = False,
    **_: typing.Any,
) -> typing.Any:
    """Example:
    ```
    class MyCtx(GlobalContext):
        __ctx_name__ = "my_ctx"

        name: str
        URL: typing.Final[str] = ctx_var(default="https://google.com", init=False, const=True)

    ctx = MyCtx(name="John")
    ctx.URL  #: 'https://google.com'
    ctx.URL = '...'  #: type checking error & exception 'TypeError'
    ```
    """
    return CtxVar(value=default, factory=default_factory, const=const)


class ConditionalLock:
    def __init__(self, lock: threading.RLock, use_lock: bool) -> None:
        self.lock = lock
        self.use_lock = use_lock

    def __enter__(self) -> typing.Self:
        if self.use_lock:
            self.lock.__enter__()
        return self

    def __exit__(self, exc_type: typing.Any, exc_val: typing.Any, exc_tb: typing.Any) -> None:
        if self.use_lock:
            self.lock.__exit__(exc_type, exc_val, exc_tb)


def runtime_init[T: ABCGlobalContext](cls: type[T], /) -> type[T]:
    r'''Initialization the global context at runtime.

    ```python
    @runtime_init
    class Box(ABCGlobalContext):
        __ctx_name__ = "box"

        cookies: list[Cookie] = ctx_var(default_factory=lambda: [ChocolateCookie()], init=False)
        """
        init=False means that when calling the class constructor it will not be necessary
        to pass this field to the class constructor, because it will already be initialized.
        """

    box = Box()  # So, this global context has already been initialized, so calling the class
                 # immediately returns an initialized instance of this class from the memory storage.

    box.cookies.append(OatmealCookie())
    print(box.cookies)  # [<ChocolateCookie>, <OatmealCookie>]
    ```
    '''
    cls()  # Init an instance of the global context.
    return cls


@dataclasses.dataclass(frozen=True, eq=False, slots=True)
class RootAttr:
    name: str
    can_be_read: bool = dataclasses.field(default=True, kw_only=True)
    can_be_rewritten: bool = dataclasses.field(default=False, kw_only=True)

    def __eq__(self, __value: str) -> bool:
        return self.name == __value


@dataclasses.dataclass(repr=False, frozen=True, slots=True)
class Storage:
    """Thread-safe storage for GlobalContext instances.

    Uses threading.RLock() for thread synchronization to ensure safe
    concurrent access to the internal storage dictionary.
    """

    _storage: dict[str | None, GlobalContext] = dataclasses.field(
        default_factory=lambda: {},
        init=False,
    )
    _lock: threading.RLock = dataclasses.field(
        default_factory=threading.RLock,
        init=False,
    )

    def __repr__(self) -> str:
        return "<{}: {}>".format(
            type(self).__name__,
            ", ".join(repr(x) for x in self._storage),
        )

    @property
    def storage(self) -> dict[str | None, GlobalContext]:
        with self._lock:
            return self._storage.copy()

    def set(self, name: str | None, context: GlobalContext) -> None:
        with self._lock:
            self._storage.setdefault(name, context)

    def get(self, ctx_name: str) -> Option[GlobalContext]:
        with self._lock:
            return from_optional(self._storage.get(ctx_name))

    def delete(self, ctx_name: str) -> None:
        with self._lock:
            assert self._storage.pop(ctx_name, None) is not None, f"Context {ctx_name!r} is not defined in storage."


@typing.dataclass_transform(
    kw_only_default=True,
    order_default=True,
    frozen_default=False,
    field_specifiers=(ctx_var,),
)
class GlobalContext[CtxValueT = typing.Any](ABCGlobalContext, dict[str, GlobalCtxVar[CtxValueT]]):
    """Global context class with optional thread safety.

    `GlobalContext` is a dictionary with additional methods for working with context.
    Thread safety can be enabled via the `thread_safe` parameter for concurrent access.

    Examples:
    ```
    # Basic usage (not thread-safe, better performance)
    ctx = GlobalContext("my_ctx")
    ctx["client"] = Client()

    # Thread-safe usage
    ctx = GlobalContext("my_ctx", thread_safe=True)
    ctx.host = CtxVar("128.0.0.7:8888", const=True)

    # Thread-safe subclass
    class MyContext(GlobalContext, thread_safe=True):
        __ctx_name__ = "my_ctx"

        friend: str

    ctx = MyContext()  # thread-safe
    ```

    """

    __ctx_name__: str | None
    """Global context name."""

    __thread_safe__: bool
    """Whether this context instance uses thread synchronization."""

    __storage__: typing.ClassVar[Storage] = Storage()
    """Storage memory; this is the storage where all initialized global contexts are stored."""

    __context_lock__: typing.ClassVar[threading.RLock] = threading.RLock()
    """Class-level lock for thread-safe GlobalContext operations."""

    __root_attributes__: typing.ClassVar[tuple[RootAttr, ...]] = (
        RootAttr(name="__ctx_name__"),
        RootAttr(name="__root_attributes__"),
        RootAttr(name="__storage__"),
        RootAttr(name="__context_lock__"),
    )
    """The sequence of root attributes of this class including this attribute."""

    def __init_subclass__(cls, *, thread_safe: bool = False, **kwargs: typing.Any) -> None:
        super().__init_subclass__(**kwargs)
        cls.__class_thread_safe__ = thread_safe

    def __new__(
        cls,
        ctx_name: str | None = None,
        /,
        *,
        thread_safe: bool | None = None,
        **variables: typing.Any | CtxVar[CtxValueT],
    ) -> typing.Self:
        """Create or get from storage a new `GlobalContext` object with optional thread safety."""
        # Determine thread_safe setting: explicit parameter > class setting > default False
        if thread_safe is None:
            thread_safe = getattr(cls, "__class_thread_safe__", False)

        with cls.__context_lock__:
            if cls is not GlobalContext:
                defaults = {}
                for name in cls.__annotations__:
                    if name in cls.__dict__ and name not in cls.__root_attributes__:
                        defaults[name] = getattr(cls, name)
                        delattr(cls, name)

                        default_ = defaults[name]
                        if isinstance(default_, CtxVar) and default_.const:
                            variables.pop(name, None)

                variables = defaults | variables

            ctx_name = getattr(cls, "__ctx_name__", ctx_name)
            if (ctx := cls.__storage__.storage.get(ctx_name)) is None:
                ctx = dict.__new__(cls, ctx_name)
                cls.__storage__.set(ctx_name, ctx)

        # Set thread_safe attribute outside the critical section
        ctx.__thread_safe__ = True if ctx_name is None else bool(thread_safe)
        ctx.set_context_variables(variables)
        return ctx  # type: ignore

    def __init__(
        self,
        ctx_name: str | None = None,
        /,
        *,
        thread_safe: bool | None = None,
        **variables: CtxValueT | CtxVariable[CtxValueT],
    ) -> None:
        if not hasattr(self, "__ctx_name__"):
            self.__ctx_name__ = ctx_name

        if not hasattr(self, "__thread_safe__"):
            if thread_safe is None:
                thread_safe = getattr(self.__class__, "__class_thread_safe__", False)
            self.__thread_safe__ = bool(thread_safe)

        if variables and not self:
            self.set_context_variables(variables)

    def __repr__(self) -> str:
        return "<{}: {{ {} }}>".format(
            f"{fullname(self)}@{self.ctx_name}{' (thread-safe)' if self.thread_safe else ''}",
            ", ".join(var_name for var_name in self),
        )

    def __eq__(self, __value: object) -> bool:
        """Returns True if the names of context stores
        that use self and __value instances are equivalent.
        """
        if not isinstance(__value, type(self)):
            return NotImplemented
        return self.__ctx_name__ == __value.__ctx_name__ and self == __value

    def __setitem__(self, __name: str, __value: CtxValueT | CtxVariable[CtxValueT]) -> None:
        with self.lock_context:
            if is_dunder(__name):
                raise NameError("Cannot set a context variable with a dunder name.")

            var = self.get(__name)
            if var and (var.value.const and var.value.value is not NOVALUE):
                raise TypeError(f"Unable to set variable {__name!r}, because it's a constant.")

            dict.__setitem__(
                self,
                __name,
                GlobalCtxVar.from_var(
                    name=__name,
                    ctx_value=__value,
                    const=var.map(lambda var: var.const).unwrap_or(False),
                ),
            )

    def __getitem__(self, __name: str) -> CtxValueT:
        with self.lock_context:
            value = self.get(__name).unwrap().value
            if value is NOVALUE:
                raise NameError(f"Variable {__name!r} is not defined in {self.ctx_name!r}.")
            return value

    def __delitem__(self, __name: str) -> None:
        with self.lock_context:
            var = self.get(__name).unwrap()
            if var.const:
                raise TypeError(f"Unable to delete variable {__name!r}, because it's a constant.")

            if var.value is NOVALUE:
                raise NameError(f"Variable {__name!r} is not defined in {self.ctx_name!r}.")

            dict.__delitem__(self, __name)

    @root_protection
    def __setattr__(self, __name: str, __value: CtxValueT | CtxVariable[CtxValueT]) -> None:
        """Setting a context variable."""
        if is_dunder(__name):
            return object.__setattr__(self, __name, __value)
        self.__setitem__(__name, __value)

    @root_protection
    def __getattr__(self, __name: str) -> CtxValueT:
        """Getting a context variable."""
        if is_dunder(__name):
            return object.__getattribute__(self, __name)
        return self.__getitem__(__name)

    @root_protection
    def __delattr__(self, __name: str) -> None:
        """Removing a context variable."""
        if is_dunder(__name):
            return object.__delattr__(self, __name)
        self.__delitem__(__name)

    @property
    def ctx_name(self) -> str:
        """Global context name."""
        return self.__ctx_name__ or "<anonymous global context at %#x>" % id(self)

    @property
    def thread_safe(self) -> bool:
        return self.__thread_safe__

    @cached_property
    def lock_context(self) -> ConditionalLock:
        """Conditional lock context manager based on thread_safe setting."""
        return ConditionalLock(type(self).__context_lock__, self.__thread_safe__)

    @classmethod
    def is_root_attribute(cls, name: str) -> bool:
        """Returns True if name is a root attribute."""
        return name in cls.__root_attributes__

    def setdefault_value(self, name: str, default: CtxValueT) -> CtxValueT:
        with self.lock_context:
            if name in self:
                return self[name]
            self[name] = default
            return default

    def set_context_variables(self, variables: typing.Mapping[str, CtxValueT | CtxVariable[CtxValueT]]) -> None:
        """Set context variables from mapping."""
        with self.lock_context:
            for name, var in variables.items():
                if is_dunder(name):
                    raise NameError("Cannot set a context variable with a dunder name.")

                existing_var = self.get(name)
                if existing_var and (existing_var.value.const and existing_var.value.value is not NOVALUE):
                    raise TypeError(f"Unable to set variable {name!r}, because it's a constant.")

                dict.__setitem__(
                    self,
                    name,
                    GlobalCtxVar.from_var(
                        name=name,
                        ctx_value=var,
                        const=existing_var.map(lambda v: v.const).unwrap_or(False),
                    ),
                )

    def get_root_attribute(self, name: str) -> Option[RootAttr]:
        """Get root attribute by name."""
        if self.is_root_attribute(name):
            for rattr in self.__root_attributes__:
                if rattr.name == name:
                    return Some(rattr)
        return Nothing()

    def items(self) -> list[tuple[str, GlobalCtxVar[CtxValueT]]]:
        """Return context variables as set-like items."""
        return list(dict.items(self))

    def keys(self) -> list[str]:
        """Returns context variable names as keys."""
        return list(dict.keys(self))

    def values(self) -> list[GlobalCtxVar[CtxValueT]]:
        """Returns context variables as values."""
        return list(dict.values(self))

    def update(self, other: typing.Self) -> None:
        with self.lock_context:
            dict.update(self, dict(other.items()))

    def copy(self) -> typing.Self:
        """Copy context. Returns copied context without ctx_name."""
        copied_ctx = self.__class__(thread_safe=self.thread_safe)
        copied_ctx.set_context_variables({name: var.value for name, var in self.dict().items()})
        return copied_ctx

    def dict(self) -> dict[str, GlobalCtxVar[CtxValueT]]:
        """Returns context as dict."""
        return {name: deepcopy(var) for name, var in self.items()}  # type: ignore

    @typing.overload
    def pop(self, var_name: str) -> Option[GlobalCtxVar[CtxValueT]]: ...

    @typing.overload
    def pop[T = typing.Any](
        self,
        var_name: str,
        var_value_type: type[T],
    ) -> Option[GlobalCtxVar[T]]: ...

    def pop(self, var_name: str, var_value_type: typing.Any = object) -> typing.Any:
        with self.lock_context:
            val = self.get(var_name, var_value_type)

            if val:
                var = dict.get(self, var_name)
                if var and var.const:
                    raise TypeError(f"Unable to delete variable {var_name!r}, because it's a constant.")
                if var and var.value is NOVALUE:
                    raise NameError(f"Variable {var_name!r} is not defined in {self.ctx_name!r}.")
                dict.__delitem__(self, var_name)
                return val

            return Nothing()

    @typing.overload
    def get(self, var_name: str) -> Option[GlobalCtxVar[CtxValueT]]: ...

    @typing.overload
    def get[T = typing.Any](
        self,
        var_name: str,
        var_value_type: type[T],
    ) -> Option[GlobalCtxVar[T]]: ...

    def get(self, var_name: str, var_value_type: typing.Any = object) -> typing.Any:
        """Get context variable by name."""
        var_value_type = typing.Any if var_value_type is object else var_value_type
        generic_types = typing.get_args(get_orig_class(self))

        if generic_types and var_value_type is object:
            var_value_type = generic_types[0]

        var = dict.get(self, var_name)
        if var is None:
            return Nothing()

        assert type_check(var.value, var_value_type), (
            "Context variable value type of {!r} does not correspond to the expected type {!r}.".format(
                type(var.value).__name__,
                (getattr(var_value_type, "__name__") if isinstance(var_value_type, type) else repr(var_value_type)),
            )
        )
        return Some(var)

    @typing.overload
    def get_value(self, var_name: str) -> Option[CtxValueT]: ...

    @typing.overload
    def get_value[T = typing.Any](
        self,
        var_name: str,
        var_value_type: type[T],
    ) -> Option[T]: ...

    def get_value(self, var_name: str, var_value_type: typing.Any = object) -> typing.Any:
        """Get context variable value by name."""
        return self.get(var_name, var_value_type).map(lambda var: var.value)

    def rename(self, old_var_name: str, new_var_name: str) -> Result[_, str]:
        with self.lock_context:
            var = self.get(old_var_name).unwrap()
            if var.const:
                return Error(f"Unable to rename variable {old_var_name!r}, because it's a constant.")

            # Atomic rename operation
            dict.__delitem__(self, old_var_name)
            dict.__setitem__(
                self,
                new_var_name,
                GlobalCtxVar.from_var(
                    name=new_var_name,
                    ctx_value=var.value,
                    const=var.const,
                ),
            )
            return Ok(_())

    def clear(self, *, include_consts: bool = False) -> None:
        """Clear context. If `include_consts = True`, the context is fully cleared."""
        with self.lock_context:
            if not self:
                return

            if include_consts:
                logger.warning(
                    "Constants from the global context {!r} have been cleaned up!",
                    self.ctx_name + " at %#x" % id(self),
                )
                return dict.clear(self)

            items_to_delete = []
            for name, var in dict.items(self):
                if not var.const:
                    items_to_delete.append(name)

            for name in items_to_delete:
                dict.__delitem__(self, name)

    def delete_ctx(self) -> Result[_, str]:
        """Delete context by `ctx_name`."""
        with self.lock_context:
            if not self.__ctx_name__:
                return Error("Cannot delete unnamed context.")

            ctx = self.__storage__.get(self.ctx_name).unwrap()
            dict.clear(ctx)
            self.__storage__.delete(self.ctx_name)

            logger.warning(f"Global context {self.ctx_name!r} has been deleted!")
            return Ok(_())


__all__ = (
    "ABCGlobalContext",
    "CtxVar",
    "CtxVariable",
    "GlobalContext",
    "GlobalCtxVar",
    "RootAttr",
    "Storage",
    "ctx_var",
    "runtime_init",
)



import asyncio
import contextlib
import datetime
import enum
import signal
import typing
from contextlib import suppress

from telegrinder.modules import logger
from telegrinder.tools.aio import TaskGroup, cancel_future, loop_is_running, run_task
from telegrinder.tools.final import Final
from telegrinder.tools.fullname import fullname
from telegrinder.tools.lifespan import (
    CoroutineFunc,
    CoroutineTask,
    DelayedTask,
    Lifespan,
    Task,
    to_coroutine_task,
)
from telegrinder.tools.singleton.singleton import Singleton

if typing.TYPE_CHECKING:
    from contextvars import Context

    type Tasks = set[asyncio.Task[typing.Any]]
    type DelayedFunctionDecorator[**P, R] = typing.Callable[[typing.Callable[P, R]], DelayedFunction[P, R]]

    class DelayedFunction[**P, R](typing.Protocol):
        __name__: str
        __delayed_task__: DelayedTask[typing.Callable[P, CoroutineTask[R]]]

        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> CoroutineTask[R]: ...

        def cancel(self) -> bool:
            """Cancel delayed task."""
            ...


class Timer(datetime.timedelta):
    repeat = False

    def __call__[**P, R](self, function: CoroutineFunc[P, R], /) -> DelayedFunction[P, R]:
        loop_wrapper = LoopWrapper()
        loop_wrapper.add_task(DelayedTask(function, seconds=self.total_seconds(), repeat=self.repeat))
        return function  # type: ignore


class Interval(Timer):
    repeat = True


@enum.unique
class LoopWrapperState(enum.Enum):
    NOT_RUNNING = enum.auto()
    RUNNING = enum.auto()
    RUNNING_MANUALLY = enum.auto()
    SHUTDOWN = enum.auto()


@typing.final
class LoopWrapper(Singleton, Final):
    _loop: asyncio.AbstractEventLoop
    _lifespan: Lifespan
    _future_tasks: list[CoroutineTask[typing.Any]]
    _state: LoopWrapperState
    _all_tasks: set[asyncio.Task[typing.Any]]
    _event_stop: asyncio.Event
    _run_lw_task: asyncio.Handle
    _is_attached_to_running_loop: bool

    timer = Timer
    interval = Interval

    __slots__ = (
        "_lifespan",
        "_future_tasks",
        "_state",
        "_all_tasks",
        "_loop",
        "_event_stop",
        "_run_lw_task",
        "_is_attached_to_running_loop",
    )

    def __init__(self) -> None:
        try:
            self._loop = asyncio.get_event_loop()
        except RuntimeError:
            self._loop = asyncio.new_event_loop()

        self._run_lw_task = self._run_lw_later()
        self._lifespan = Lifespan()
        self._event_stop = asyncio.Event()
        self._state = LoopWrapperState.NOT_RUNNING
        self._future_tasks = [self._waiter_stop()]
        self._all_tasks = set()
        self._is_attached_to_running_loop = False

        signal.signal(signal.SIGTERM, lambda *_: self.stop())

    def __repr__(self) -> str:
        return "<{}: loop={!r}, lifespan={!r}>".format(
            fullname(self) + (" (running)" if self.running else ""),
            self._loop,
            self._lifespan,
        )

    def __call__(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def lifespan(self) -> Lifespan:
        return self._lifespan

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def time(self) -> float:
        return self._loop.time()

    @property
    def running(self) -> bool:
        return self._state in {LoopWrapperState.RUNNING, LoopWrapperState.RUNNING_MANUALLY}

    @property
    def shutdown(self) -> bool:
        return self._state is LoopWrapperState.SHUTDOWN

    async def _run_async_event_loop(self) -> None:
        if not self.running:
            self._state = LoopWrapperState.RUNNING
            async with self._async_wrap_loop():
                await self._run()

    async def _run(self) -> None:
        await logger.adebug("Running loop wrapper")

        while self._future_tasks:
            self._create_task(self._future_tasks.pop(0))

        while self.running and (tasks := self._get_all_tasks()):
            tasks_results, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
            for task_result in tasks_results:
                try:
                    task_result.result()
                except Exception:
                    await logger.aexception("Traceback message below:")

    async def _cancel_tasks(self) -> None:
        with suppress(asyncio.exceptions.CancelledError):
            await cancel_future(asyncio.gather(*self._get_all_tasks(), return_exceptions=True))

    @contextlib.asynccontextmanager
    async def _async_wrap_loop(self) -> typing.AsyncGenerator[typing.Any, None]:
        try:
            try:
                await self._lifespan._start()
            finally:
                yield
        except asyncio.CancelledError:
            await logger.adebug("Cancelling tasks...")
            await self._cancel_tasks()
        finally:
            await self._shutdown()

    async def _shutdown(self) -> None:
        await self.lifespan._shutdown(None, None, None)
        await logger.adebug("Shutting down loop wrapper")
        self._state = LoopWrapperState.SHUTDOWN

    async def _waiter_stop(self) -> None:
        await self._event_stop.wait()
        self._state = LoopWrapperState.SHUTDOWN
        self._event_stop.clear()
        await self._shutdown()
        await self._cancel_tasks()

    def _run_lw_later(self) -> asyncio.Handle:
        return self._loop.call_soon_threadsafe(lambda l: l.create_task(self._run_async_event_loop()), self._loop)

    def _create_task(
        self,
        coro: CoroutineTask[typing.Any],
        /,
        name: str | None = None,
        context: Context | None = None,
    ) -> asyncio.Task[typing.Any]:
        task = self._loop.create_task(coro, name=name, context=context)
        self._all_tasks.add(task)
        task.add_done_callback(self._all_tasks.discard)

        try:
            return task
        finally:
            del task

    def _get_all_tasks(self) -> Tasks:
        """Get a set of all tasks from the loop wrapper and event loop (`exclude the current task if any`)."""

        return (self._all_tasks | asyncio.all_tasks(loop=self._loop)).symmetric_difference(
            set() if (task := asyncio.current_task(self._loop)) is None else {task},
        )

    def _close_loop(self) -> None:
        if not self._loop.is_closed():
            logger.debug("Closing event loop {!r}", self._loop)
            self._loop.close()

    @contextlib.contextmanager
    def _wrap_loop(self, *, close_loop: bool = True) -> typing.Generator[typing.Any, None, None]:
        try:
            try:
                self.lifespan.start(self._loop)
            finally:
                yield
        except (KeyboardInterrupt, SystemExit) as e:
            print(flush=True)
            logger.info(f"Caught {e.__class__.__name__}, cancelling tasks")
            run_task(self._cancel_tasks(), loop=self._loop)

            if isinstance(e, SystemExit):
                raise
        finally:
            run_task(self._shutdown(), loop=self._loop)

            if close_loop:
                self._close_loop()

    def run(self, *, close_loop: bool = True) -> typing.NoReturn:  # type: ignore
        if self.running:
            raise RuntimeError("Loop wrapper already running.")

        self._state = LoopWrapperState.RUNNING_MANUALLY
        with self._wrap_loop(close_loop=close_loop):
            run_task(self._run(), loop=self._loop)

    def stop(self) -> None:
        if not self._event_stop.is_set():
            self._event_stop.set()

    def add_task(self, task: Task[..., typing.Any], /) -> None:
        coro_task = to_coroutine_task(task)

        if not self.running:
            self._future_tasks.append(coro_task)

            if self._is_attached_to_running_loop is False and loop_is_running():
                self.attach_to_running_loop()
        else:
            self._create_task(coro_task)

    def attach_to_running_loop(self) -> None:
        if self.running:
            raise RuntimeError("Cannot attach the running loop wrapper to the running loop.")

        self._loop = asyncio.get_running_loop()
        self._run_lw_task.cancel()
        self._is_attached_to_running_loop = True
        self._run_lw_task = self._run_lw_later()

    def create_task_group[T](self) -> TaskGroup[typing.Any]:
        loop = asyncio.get_running_loop()
        if not self.running and self._loop is not loop:
            self.attach_to_running_loop()
        return TaskGroup(loop)


__all__ = ("DelayedTask", "LoopWrapper", "to_coroutine_task")



import typing

if typing.TYPE_CHECKING:
    from telegrinder.tools.serialization.abc import ABCDataSerializer

IDENT_KEY: typing.Final = "__key__"
SERIALIZER_KEY: typing.Final = "__serializer__"


def get_model_ident_key(model: typing.Any, /) -> str | None:
    return getattr(model, IDENT_KEY, None)


def get_model_serializer(model: typing.Any, /) -> type[ABCDataSerializer[typing.Any]] | None:
    return getattr(model, SERIALIZER_KEY, None)


__all__ = ("get_model_ident_key", "get_model_serializer")



from nodnod.error import NodeError
from nodnod.interface.scalar import scalar_node
from nodnod.node import Scalar

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.types.enums import ChatType
from telegrinder.types.objects import Chat, MessageOriginChannel

type MessageChannelPost = MessageOriginChannel
type Post = MessageCute
type PostId = int


@scalar_node
class ChatMessageChannelPost:
    @classmethod
    def __compose__(cls, message: MessageCute) -> MessageChannelPost:
        forward_origin = message.forward_origin.expect(NodeError("Message has no forward origin."))
        return forward_origin.only(MessageOriginChannel).expect(NodeError("Message forward origin is not a channel."))


@scalar_node
class ChatMessageChannelPostId:
    @classmethod
    def __compose__(cls, message_channel_post: ChatMessageChannelPost) -> PostId:
        return message_channel_post.message_id


@scalar_node
class ChatMessageChannelPostChannel:
    @classmethod
    def __compose__(cls, message_channel_post: ChatMessageChannelPost) -> Channel:
        return message_channel_post.chat


@scalar_node
class ChatMessageChannelPostChannelId:
    @classmethod
    def __compose__(cls, message_channel_post: ChatMessageChannelPost) -> ChannelId:
        return message_channel_post.chat.id


@scalar_node
class ChatMessageChannelPostAuthor:
    @classmethod
    def __compose__(cls, message_channel_post: ChatMessageChannelPost) -> str:
        return message_channel_post.author_signature.expect(
            NodeError("Discussion has no signature of the post author."),
        )


@scalar_node
class ChannelPostNode:
    @classmethod
    def __compose__(cls, message: MessageCute) -> ChannelPost:
        if message.chat.type != ChatType.CHANNEL:
            raise NodeError("Message is not a channel post.")
        return message


@scalar_node
class ChannelPostId:
    @classmethod
    def __compose__(cls, channel_post: ChannelPostNode) -> PostId:
        return channel_post.message_id


@scalar_node
class ChannelNode:
    @classmethod
    def __compose__(cls, channel_post: ChannelPostNode) -> Channel:
        return channel_post.chat


@scalar_node
class ChannelIdNode:
    @classmethod
    def __compose__(cls, channel_post: ChannelPostNode) -> ChannelId:
        return channel_post.chat.id


type Channel = Scalar[Chat, ChannelNode]
type ChannelId = Scalar[int, ChannelIdNode]
type ChannelPost = Scalar[MessageCute, ChannelPostNode]


__all__ = (
    "Channel",
    "ChannelId",
    "ChannelPost",
    "ChannelPostId",
    "ChatMessageChannelPost",
    "ChatMessageChannelPostAuthor",
    "ChatMessageChannelPostChannel",
    "ChatMessageChannelPostChannelId",
    "ChatMessageChannelPostId",
)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import PollAnswer


class PollAnswerCute(BaseCute[PollAnswer], PollAnswer, kw_only=True):
    pass


__all__ = ("PollAnswerCute",)



from telegrinder.bot.cute_types.inline_query import InlineQueryCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.markup import Markup, PatternLike, check_string
from telegrinder.types.enums import ChatType

type InlineQuery = InlineQueryCute


class HasLocation(ABCRule):
    def check(self, query: InlineQuery) -> bool:
        return bool(query.location)


class InlineQueryChatType(ABCRule):
    def __init__(self, chat_type: ChatType, /) -> None:
        self.chat_type = chat_type

    def check(self, query: InlineQuery) -> bool:
        return query.chat_type.map(lambda x: x == self.chat_type).unwrap_or(False)


class InlineQueryText(ABCRule):
    def __init__(self, texts: str | list[str], *, lower_case: bool = False) -> None:
        self.texts = {text.lower() if lower_case else text for text in ([texts] if isinstance(texts, str) else texts)}
        self.lower_case = lower_case

    def check(self, query: InlineQuery) -> bool:
        return (query.query.lower() if self.lower_case else query.query) in self.texts


class InlineQueryMarkup(ABCRule):
    def __init__(self, patterns: PatternLike | list[PatternLike], /) -> None:
        self.patterns = Markup(patterns).patterns

    def check(self, query: InlineQuery, ctx: Context) -> bool:
        return check_string(self.patterns, query.query, ctx)


__all__ = (
    "HasLocation",
    "InlineQueryChatType",
    "InlineQueryMarkup",
    "InlineQueryText",
)



import importlib.util as importlib_util
import os
import pathlib
import sys
import typing
from abc import ABC, abstractmethod

from telegrinder.api.api import API
from telegrinder.types.objects import Update


class PathExistsError(BaseException):
    pass


class ABCDispatch(ABC):
    @abstractmethod
    async def feed(self, api: API, update: Update) -> typing.Any:
        pass

    @abstractmethod
    def load(self, external: typing.Self) -> None:
        pass

    def load_many(self, *externals: typing.Self) -> None:
        for external in externals:
            self.load(external)

    def load_from_dir(
        self,
        directory: str | pathlib.Path,
        *,
        recursive: bool = False,
    ) -> bool:
        """Loads dispatchers from a directory containing Python modules where global variables
        are declared with instances of dispatch.
        Returns True if dispatchers were found, otherwise False.
        """

        directory = pathlib.Path(directory)

        if not directory.exists():
            raise PathExistsError(f"Path `{directory!s}` doesn't exists.")

        found = False
        files_iter = os.walk(directory) if recursive else [(directory, [], os.listdir(directory))]

        for root, _, files in files_iter:
            for f in files:
                if f.endswith(".py") and f != "__init__.py":
                    module_path = os.path.join(root, f)
                    relative_path = os.path.relpath(module_path, sys.path[0])
                    module_name = os.path.splitext(relative_path)[0].replace(os.sep, ".")

                    if module_name not in sys.modules:
                        spec = importlib_util.spec_from_file_location(module_name, module_path)
                        if spec is None or spec.loader is None:
                            continue

                        module = importlib_util.module_from_spec(spec)
                        sys.modules[module_name] = module
                        spec.loader.exec_module(module)
                    else:
                        module = sys.modules[module_name]

                    for obj in vars(module).values():
                        if isinstance(obj, type(self)):
                            found = True
                            self.load(obj)

        return found


__all__ = ("ABCDispatch",)



import dataclasses
import typing

from nodnod.error import NodeError
from nodnod.interface.generic import generic_node
from nodnod.node import Node

from telegrinder.bot.dispatch.context import Context

type ExceptionType = type[Exception]


def can_catch[ExceptionT: Exception](
    exc: Exception | ExceptionType,
    exc_types: type[ExceptionT] | tuple[type[ExceptionT], ...],
) -> typing.TypeGuard[ExceptionT]:
    return issubclass(exc, exc_types) if isinstance(exc, type) else isinstance(exc, exc_types)


@generic_node
@dataclasses.dataclass(kw_only=True, frozen=True)
class Error[*Exceptions = *tuple[type[Exception], ...]](Node):
    exception_update: Exception

    @property
    def exception[T: Exception = Exception](self: Error[*tuple[T, ...]]) -> T:
        return self.exception_update  # type: ignore[reportReturnType]

    @classmethod
    def __compose__(
        cls,
        exceptions: tuple[typing.Unpack[Exceptions]],
        context: Context,
    ) -> typing.Self:
        exception_update = context.exception_update.expect(NodeError("No exception."))

        if can_catch(exception_update, exceptions):  # type: ignore
            return cls(exception_update=exception_update)

        raise NodeError("Foreign exception.")


__all__ = ("Error",)



"""Specific `scopes` for node scope system.

Scopes:
- `GLOBAL`: compose only once during runtime, and later be stored and reused when needed ~ `@global_node`
- `PER_EVENT`: compose once per event, so if during the composition the node was already composed, it will be reused and won't be composed twice ~ `@per_event`
- `PER_CALL`: compose each time any node will require it to build itself or if we require it to be delivered into the handler ~ `@per_call`
"""

from __future__ import annotations

import enum
import typing

from nodnod.node import Node
from nodnod.scope import Scope

from telegrinder.modules import logger
from telegrinder.node.utils import as_node
from telegrinder.tools.global_context.builtin_context import TelegrinderContext

if typing.TYPE_CHECKING:
    from telegrinder.node.scope import NodeScope

    # Import members from NodeScope
    GLOBAL, PER_CALL, PER_EVENT = NodeScope.GLOBAL, NodeScope.PER_CALL, NodeScope.PER_EVENT

TELEGRINDER_CONTEXT: typing.Final = TelegrinderContext()


@TELEGRINDER_CONTEXT.loop_wrapper.lifespan.on_shutdown
async def close_node_global_scope() -> None:
    logger.debug("Closing node global scope")

    try:
        await TELEGRINDER_CONTEXT.close_global_scope()
    except Exception as error:
        logger.error("While closing node global scope, an error occurred: {!r}", error)


class NodeScopeInfo(dict[type[Node], "NodeScope"]):
    def set_node_scope[T: Node[typing.Any, typing.Any]](self, node: type[T], scope: NodeScope, /) -> None:
        self[node] = scope

    def get_node_scope[T: Node[typing.Any, typing.Any]](self, node: type[T], /) -> NodeScope | None:
        return self.get(node, None)


NODE_SCOPE_INFO: typing.Final = NodeScopeInfo()


class MappedScopes(dict[type[Node], Scope]):
    scopes: dict[NodeScope, Scope]

    def __init__(self, global_scope: Scope, per_event_scope: Scope) -> None:
        self.scopes = {NodeScope.GLOBAL: global_scope, NodeScope.PER_EVENT: per_event_scope}

    def get(self, key: type[Node], default: Scope | None = None) -> Scope | None:
        scope = NODE_SCOPE_INFO.get_node_scope(key) or NodeScope.PER_EVENT
        return self.scopes.get(scope, default)


# Declare NodeScope members in a global scope
@enum.global_enum
class NodeScope(enum.StrEnum):
    GLOBAL = "global"
    PER_CALL = "local"
    PER_EVENT = "event"

    def __call__[T](self, node: type[T], /) -> type[T]:
        NODE_SCOPE_INFO.set_node_scope(as_node(node), self)
        return node


# Decorators
global_node, per_call, per_event = GLOBAL, PER_CALL, PER_EVENT


__all__ = (
    "GLOBAL",
    "NODE_SCOPE_INFO",
    "PER_CALL",
    "PER_EVENT",
    "MappedScopes",
    "NodeScope",
    "global_node",
    "per_call",
    "per_event",
)



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.rules import HasEntities, IsChat, IsPrivate, MessageEntities
from telegrinder.tools.formatting import HTML, bold, mention
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import MessageEntity

api = API(Token.from_env())
bot = Telegrinder(api)


@bot.on.message(IsChat(), MessageEntities(MessageEntityType.MENTION))
async def handler_mention_me(message: Message, message_entities: list[MessageEntity]):
    my_username = (await api.get_me()).map(lambda me: me.username).unwrap()
    if (
        not my_username
        or my_username != message.text.unwrap()[message_entities[0].offset + 1 : message_entities[0].length]
    ):
        return

    await message.delete()
    await message.answer(
        HTML
        << bold(mention(message.from_user.first_name, user_id=message.from_user.id))
        << " don't mention me please!",
        parse_mode=HTML.PARSE_MODE,
    )


@bot.on.message(IsPrivate(), HasEntities())
async def handler_entities(message: Message):
    await message.answer(
        "Nice {}: {}!".format(
            "entities" if len(message.entities.unwrap()) > 1 else "entity",
            ", ".join(map(lambda e: e.type, message.entities.unwrap())),
        )
    )


bot.run_forever()



import abc
import typing
from functools import cached_property

import msgspec
from kungfu.library.monad.result import Result

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance

type ModelType = msgspec.Struct | DataclassInstance


class ABCDataSerializer[Data = typing.Any](abc.ABC):
    ident_key: str | None = None

    @abc.abstractmethod
    def __init__(self, data_type: type[Data], /) -> None:
        pass

    @cached_property
    def key(self) -> str:
        return self.ident_key + "_" if self.ident_key else ""

    @abc.abstractmethod
    def serialize(self, data: Data) -> str:
        pass

    @abc.abstractmethod
    def deserialize(self, serialized_data: str) -> Result[Data, str]:
        pass


__all__ = ("ABCDataSerializer",)



from .config import (
    Config,
    TypedDefaultParameter,
    dec_hook,
)
from .schema import (
    MethodParameter,
    MethodSchema,
    Model,
    ObjectField,
    ObjectSchema,
    TelegramBotAPISchema,
)
from .types.methods import *
from .types.objects import *

__all__ = (
    "Config",
    "GeneratorMethods",
    "GeneratorObjects",
    "MethodParameter",
    "MethodSchema",
    "MethodsParams",
    "MethodsParamsAnnotations",
    "MethodsParamsAnnotationsAnnotations",
    "MethodsParamsAnnotationsAnnotationsParam",
    "MethodsParamsLiteralTypes",
    "MethodsParamsLiteralTypesParam",
    "MethodsParamsLiteralTypesParam",
    "Model",
    "ObjectField",
    "ObjectSchema",
    "ObjectsFields",
    "ObjectsFieldsAnnotations",
    "ObjectsFieldsAnnotationsAnnotations",
    "ObjectsFieldsAnnotationsAnnotationsField",
    "ObjectsFieldsDefaults",
    "ObjectsFieldsIdByDefault",
    "ObjectsFieldsIdByDefaultField",
    "ObjectsFieldsLiteralTypes",
    "ObjectsFieldsLiteralTypesField",
    "TelegramBotAPISchema",
    "TypedDefaultParameter",
    "dec_hook",
)



import typing
from collections import deque
from functools import cached_property

from nodnod.interface.inject import inject_internals

from telegrinder.api.api import API
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware, run_post_middleware, run_pre_middleware
from telegrinder.bot.dispatch.middleware.box import MiddlewareBox
from telegrinder.bot.dispatch.router.base import Router
from telegrinder.bot.dispatch.view.base import ErrorView, View
from telegrinder.bot.dispatch.view.box import ViewBox
from telegrinder.modules import logger
from telegrinder.node.scope import PER_EVENT
from telegrinder.tools.fullname import fullname
from telegrinder.tools.global_context import TelegrinderContext
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from vbml.patcher.abc import ABCPatcher

    from telegrinder.bot.dispatch.view.base import EventView, RawEventView, View
    from telegrinder.bot.dispatch.view.media_group import MediaGroupView

NANOSECONDS_PER_MILLISECOND: typing.Final = 1_000_000_000


class ViewGetter:
    main_router: Router

    def __getattr__(self, name: str, /) -> typing.Any:
        if name in ViewBox.__dataclass_fields__ or name in ViewBox.__dict__:
            return getattr(self.main_router, name)
        return super().__getattribute__(name)


class Dispatch[
    T: MiddlewareBox = MiddlewareBox,
    ErrorHandler: ErrorView = ErrorView,
    MessageView: EventView = EventView,
    EditedMessageView: EventView = EventView,
    ChannelPostView: EventView = EventView,
    EditedChannelPostView: EventView = EventView,
    BusinessConnectionView: EventView = EventView,
    BusinessMessageView: EventView = EventView,
    EditedBusinessMessageView: EventView = EventView,
    DeletedBusinessMessagesView: EventView = EventView,
    MessageReactionView: EventView = EventView,
    MessageReactionCountView: EventView = EventView,
    InlineQueryView: EventView = EventView,
    ChosenInlineResultView: EventView = EventView,
    CallbackQueryView: EventView = EventView,
    ShippingQueryView: EventView = EventView,
    PreCheckoutQueryView: EventView = EventView,
    PurchasedPaidMediaView: EventView = EventView,
    PollView: EventView = EventView,
    PollAnswerView: EventView = EventView,
    MyChatMemberView: EventView = EventView,
    ChatMemberView: EventView = EventView,
    ChatJoinRequestView: EventView = EventView,
    ChatBoostView: EventView = EventView,
    RemovedChatBoostView: EventView = EventView,
    MediaGroup: View = MediaGroupView,
    EventError: ErrorView = ErrorView,
    RawEvent: RawEventView = RawEventView,
](
    ABCDispatch,
    ViewBox[
        MessageView,
        EditedMessageView,
        ChannelPostView,
        EditedChannelPostView,
        BusinessConnectionView,
        BusinessMessageView,
        EditedBusinessMessageView,
        DeletedBusinessMessagesView,
        MessageReactionView,
        MessageReactionCountView,
        InlineQueryView,
        ChosenInlineResultView,
        CallbackQueryView,
        ShippingQueryView,
        PreCheckoutQueryView,
        PurchasedPaidMediaView,
        PollView,
        PollAnswerView,
        MyChatMemberView,
        ChatMemberView,
        ChatJoinRequestView,
        ChatBoostView,
        RemovedChatBoostView,
        MediaGroup,
        EventError,
        RawEvent,
    ]
    if typing.TYPE_CHECKING
    else ViewGetter,
):
    type MainRouter = Router[
        MessageView,
        EditedMessageView,
        ChannelPostView,
        EditedChannelPostView,
        BusinessConnectionView,
        BusinessMessageView,
        EditedBusinessMessageView,
        DeletedBusinessMessagesView,
        MessageReactionView,
        MessageReactionCountView,
        InlineQueryView,
        ChosenInlineResultView,
        CallbackQueryView,
        ShippingQueryView,
        PreCheckoutQueryView,
        PurchasedPaidMediaView,
        PollView,
        PollAnswerView,
        MyChatMemberView,
        ChatMemberView,
        ChatJoinRequestView,
        ChatBoostView,
        RemovedChatBoostView,
        MediaGroup,
        EventError,
        RawEvent,
    ]

    main_router: MainRouter
    error_handler: ErrorHandler
    middlewares: T
    _routers: deque[Router] | None = None

    @typing.overload
    def __init__(self) -> None: ...

    @typing.overload
    def __init__(self, *, router: MainRouter) -> None: ...

    @typing.overload
    def __init__(self, *, middleware_box: T) -> None: ...

    @typing.overload
    def __init__(self, *, router: MainRouter, middleware_box: T) -> None: ...

    def __init__(
        self,
        *,
        router: MainRouter | None = None,
        error_handler: ErrorHandler | None = None,
        middleware_box: MiddlewareBox | None = None,
    ) -> None:
        self.main_router = router or Router()  # type: ignore
        self.error_handler = error_handler or ErrorView()  # type: ignore
        self.global_context = TelegrinderContext()
        self.global_scope = self.global_context.node_global_scope
        self.loop_wrapper = self.global_context.loop_wrapper
        self.middlewares = self.global_context.setdefault_value("middleware_box", middleware_box or MiddlewareBox())

    def __setitem__(self, injection_type: typing.Any, injection_value: typing.Any, /) -> None:
        self.global_scope.inject(injection_type, injection_value)

    @property
    def routers(self) -> deque[Router]:
        if self._routers is None:
            self._routers = deque((self.main_router,) if self.main_router else ())  # type: ignore
        return self._routers  # type: ignore

    @cached_property
    def raw_views(self) -> tuple[View, ...]:
        return tuple(filter(None, (router.raw for router in self.routers)))

    @property
    def patcher(self) -> ABCPatcher:
        """Alias `patcher` to get a vbml patcher from the global context."""
        return self.global_context.vbml_patcher

    @property
    def register_middleware[Middleware: ABCMiddleware](self) -> typing.Callable[[type[Middleware]], type[Middleware]]:
        """Decorator to register a custom middleware in the dispatch's middleware box."""
        return self.middlewares.__call__

    async def _handle_exceptions(
        self,
        api: API,
        update: Update,
        context: Context,
        exceptions: tuple[BaseException | BaseExceptionGroup[BaseException], ...],
    ) -> None:
        unhandled_exceptions: list[BaseException] = []

        try:
            async with self.loop_wrapper.create_task_group() as task_group:
                for exception in exceptions:
                    if isinstance(exception, BaseExceptionGroup):
                        task_group.create_task(
                            self._handle_exceptions(api, update, context.copy(), exception.exceptions)
                        )
                    elif isinstance(exception, Exception):
                        task_group.create_task(
                            self.main_router.route_view(
                                self.error_handler,
                                api,
                                update,
                                context.copy().add_exception_update(exception),
                            ),
                        )
                    else:
                        unhandled_exceptions.append(exception)
        except BaseExceptionGroup as group:
            if not unhandled_exceptions:
                raise

            raise BaseExceptionGroup(
                "Unhandled exception groups:",
                [group, BaseExceptionGroup("Unhandled exceptions:", unhandled_exceptions)],
            )

        if unhandled_exceptions:
            raise BaseExceptionGroup("Unhandled exceptions:", unhandled_exceptions)

    async def _process_views(
        self,
        views: typing.Iterable[View],
        api: API,
        update: Update,
        context: Context,
    ) -> bool:
        async with self.loop_wrapper.create_task_group() as task_group:
            for view in views:
                task_group.create_task(self.main_router.route_view(view, api, update, context.copy()))

        return any(task_group.results())

    async def _process_update_exceptions(
        self,
        api: API,
        update: Update,
        context: Context,
    ) -> None:
        await logger.adebug(
            "Processing error views with exceptions [{}] for update (id={}, type={!r})",
            ", ".join(f"{type(e).__name__}" for e in context.exceptions_update.values()),
            update.update_id,
            update.update_type,
        )

        async with self.loop_wrapper.create_task_group() as task_group:
            for router, exception in context.exceptions_update.items():
                await logger.adebug(
                    "Routing exception update (id={}, type={!r}) to router `{!r}`",
                    update.update_id,
                    update.update_type,
                    router,
                )
                task_group.create_task(
                    router.route_view(
                        router.event_error,
                        api,
                        update,
                        context.copy().add_exception_update(exception),
                    ),
                )

    async def _route_update(self, api: API, update: Update, context: Context) -> bool:
        async with self.loop_wrapper.create_task_group() as task_group:
            for router in self.routers:
                await logger.adebug(
                    "Routing update (id={}, type={!r}) to router `{!r}`",
                    update.update_id,
                    update.update_type,
                    router,
                )
                task_group.create_task(router.route(api, update, context.copy()))

        return any(task_group.results())

    async def feed(self, api: API, update: Update) -> None:
        await logger.ainfo(
            "New Update(id={}, type={!r}) received by bot (id={})",
            update.update_id,
            update.update_type,
            api.id,
        )

        per_event_scope = self.global_scope.create_child(detail=PER_EVENT)
        context = Context().add_roots(api, update, per_event_scope)

        inject_internals(per_event_scope, {API: api, Update: update})

        failed = False
        middlewares = self.middlewares
        start_time = self.loop_wrapper.time

        async with per_event_scope:
            try:
                for middleware in middlewares:
                    if await run_pre_middleware(middleware, context) is not True:
                        await logger.ainfo(
                            "Update(id={}, type={!r}) processed with dispatch's pre-middleware `{}` and raised failure.",
                            update.update_id,
                            update.update_type,
                            fullname(middleware),
                        )
                        return

                if not self.routers:
                    await logger.adebug(
                        "No corresponding routers from dispatch found for update (id={}, type={!r}).",
                        update.update_id,
                        update.update_type,
                    )
                elif not await self._route_update(api, update, context) and self.raw_views:
                    await self._process_views(self.raw_views, api, update, context)

                for middleware in middlewares:
                    await run_post_middleware(middleware, context)
            except BaseException as exc:
                failed = True

                if context.exceptions_update:
                    try:
                        await self._process_update_exceptions(api, update, context)
                    except BaseExceptionGroup as group:
                        if not self.error_handler:
                            raise

                        await logger.adebug(
                            "Dispatch caught unhandled exceptions while processing update (id={}, type={!r}), "
                            "routing to error handler...",
                            update.update_id,
                            update.update_type,
                        )
                        await self._handle_exceptions(api, update, context, group.exceptions)

                    return

                if isinstance(exc, Exception) and self.error_handler:
                    await logger.adebug(
                        "Dispatch caught an exception while processing update (id={}, type={!r}), "
                        "routing to error handler...",
                        update.update_id,
                        update.update_type,
                    )
                    await self.main_router.route_view(
                        self.error_handler, api, update, context.add_exception_update(exc)
                    )
                    return

                raise
            finally:
                if not failed:
                    elapsed_time = self.loop_wrapper.time - start_time
                    elapsed_ms = elapsed_time * 1000
                    await logger.adebug(
                        "Update (id={}, type={!r}) processed in {} {} by bot (id={})",
                        update.update_id,
                        update.update_type,
                        int(elapsed_time * NANOSECONDS_PER_MILLISECOND) if elapsed_ms < 1 else int(elapsed_ms),
                        "ns" if elapsed_ms < 1 else "ms",
                        api.id,
                    )

    def load(self, external: typing.Self) -> None:
        self.routers.extend(filter(None, external.routers))
        self.error_handler.load(external.error_handler)


__all__ = ("Dispatch",)



import typing
from contextlib import suppress

import kungfu

from telegrinder.bot.rules.abc import ABCRule
from telegrinder.node import EventNode
from telegrinder.node.utils import is_node
from telegrinder.tools.member_descriptor_proxy import MemberDescriptorProxy, evaluate_operations


class Magic[R, **P = [R]](kungfu.F[R, P], ABCRule):
    def __init__(
        self,
        field: R,
        parent: typing.Any = None,
    ) -> None:
        if isinstance(field, MemberDescriptorProxy):
            owner_cls, member_name, operations = field._objclass, field._member_name, field._operations
            node = EventNode[owner_cls]
            super().__init__(lambda obj: evaluate_operations(obj, member_name, operations))  # type: ignore

        elif is_node(field):
            node = field
            super().__init__()  # type: ignore

        else:
            node = parent
            super().__init__(field)  # type: ignore

        self.node = node
        self.required_nodes = dict(magic_value=node)

    def check(self, magic_value: typing.Any) -> bool:
        with suppress(kungfu.UnwrapError):
            self(magic_value)  # type: ignore
            return True

        return False

    def new(self, f: typing.Any, /) -> typing.Any:
        return Magic(f, parent=self.node)

    if typing.TYPE_CHECKING:

        def then[T](self, g: typing.Callable[[R], T], /) -> "Magic[T, P]": ...

        def ensure(
            self,
            chk: typing.Callable[[R], bool],
            error: typing.Callable[[R], BaseException] | BaseException | str | None = None,
        ) -> "Magic[R, P]": ...

        def expect[T, Err](
            self: kungfu.F[kungfu.Result[T, Err], P],
            error: typing.Callable[[kungfu.Result[T, Err]], BaseException] | BaseException | str | None = None,
        ) -> "Magic[T, P]": ...


__all__ = ("Magic",)



import typing

from nodnod.error import NodeError
from nodnod.node import Node
from nodnod.value import Value

from telegrinder.node.scope import TELEGRINDER_CONTEXT, global_node
from telegrinder.tools.fullname import fullname

NODEFAULT: typing.Final = object()

_Unspecialized = typing.NewType("_Unspecialized", type)


@global_node
class GlobalNode[T = _Unspecialized](Node, abstract=True):
    @typing.overload
    @classmethod
    def set(cls: type[GlobalNode[_Unspecialized]], value: typing.Self, /) -> None: ...  # type: ignore

    @typing.overload
    @classmethod
    def set(cls, value: T, /) -> None: ...

    @classmethod
    def set(cls, value: T | typing.Self, /) -> None:
        TELEGRINDER_CONTEXT.node_global_scope.push(Value(cls, value))

    @typing.overload
    @classmethod
    def get(cls: type[GlobalNode[_Unspecialized]], /) -> typing.Self:  # type: ignore
        ...

    @typing.overload
    @classmethod
    def get(cls) -> T: ...

    @typing.overload
    @classmethod
    def get[Default](cls: type[GlobalNode[_Unspecialized]], *, default: Default) -> typing.Self | Default:  # type: ignore
        ...

    @typing.overload
    @classmethod
    def get[Default](cls, *, default: Default) -> T | Default: ...

    @classmethod
    def get(cls, *, default: typing.Any = NODEFAULT) -> typing.Any:
        if default is not NODEFAULT and cls not in TELEGRINDER_CONTEXT.node_global_scope:
            return default

        if (value := TELEGRINDER_CONTEXT.node_global_scope.get(cls)) is None and default is NODEFAULT:
            raise NodeError(f"Node `{fullname(cls)}` has no global value.")

        return value.value if value is not None else default

    @classmethod
    def unset(cls) -> None:
        TELEGRINDER_CONTEXT.node_global_scope.pop(cls, None)

    @classmethod
    def __compose__(cls) -> T:
        return cls.get()


__all__ = ("GlobalNode",)



import typing

from telegrinder.bot.dispatch.handler.abc import ABCHandler

if typing.TYPE_CHECKING:
    from telegrinder.bot.cute_types.base import BaseCute

type OnDrop[Event: BaseCute[typing.Any] = typing.Any] = typing.Callable[..., typing.Awaitable[typing.Any] | typing.Any]


class WaiterActions[Event: BaseCute[typing.Any] = typing.Any](typing.TypedDict):
    on_miss: typing.NotRequired[ABCHandler]
    on_drop: typing.NotRequired[OnDrop[Event]]


__all__ = ("WaiterActions",)



import typing

from kungfu.library.monad.option import Option

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.model import Model
from telegrinder.tools.magic.shortcut import shortcut
from telegrinder.types.objects import Update

def compose_method_params[Cute: BaseCute](
    params: dict[str, typing.Any],
    update: Cute,
    *,
    default_params: set[str | tuple[str, str]] | None = None,
    validators: dict[str, typing.Callable[[Cute], bool]] | None = None,
) -> dict[str, typing.Any]: ...

class BaseShortcuts[Cute: BaseCute[typing.Any] = typing.Any]:
    cute: typing.Final[Cute]

class BaseCute[T: Model = typing.Any](Model):
    api: typing.Final[API]
    """`API` bound to the cute model."""

    ctx_api: typing.Final[API]
    """Alias for `api`."""

    raw_update: typing.Final[Update]
    """`Update` bound to the cute model."""

    @classmethod
    def __compose__(cls, update: Update, context: Context) -> typing.Self: ...
    @classmethod
    def from_update(cls, update: T, bound_api: API) -> typing.Self: ...
    def bind_raw_update(self, raw_update: Update, /) -> typing.Self: ...
    def get_raw_update(self) -> Option[Update]: ...
    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]: ...
    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]: ...

__all__ = ("BaseCute", "BaseShortcuts", "compose_method_params", "shortcut")



import sys


def get_frame_module_name() -> str:
    frame = sys._getframe()

    while frame:
        if frame.f_globals.get("__name__", "").startswith("telegrinder"):
            frame = frame.f_back
        else:
            break

    return "<module>" if frame is None else frame.f_globals.get("__name__", "<module>")


__all__ = ("get_frame_module_name",)



import typing

from kungfu.library.monad.result import Error, Ok, Result

from telegrinder import Dispatch, LoopWrapper, Message, Polling
from telegrinder.api import API, Token
from telegrinder.api.error import APIError
from telegrinder.modules import setup_logger
from telegrinder.rules import Command
from telegrinder.types.objects import User


class DummyAPI(API):
    _cached_me: User | None = None

    async def get_me(self, **other: typing.Any) -> Result[User, APIError]:
        if self._cached_me is None:
            match await super().get_me(**other):
                case Ok(user):
                    self._cached_me = user
                case Error(_) as err:
                    return err
        return Ok(self._cached_me)


setup_logger()
api = DummyAPI(token=Token.from_env())
polling = Polling(api=api)
lw = LoopWrapper()
dp = Dispatch()


@dp.raw(Command("get_me"))
async def message_handler(message: Message) -> None:
    me = (await api.get_me()).unwrap()
    await message.reply(
        "Hello, my name is {}, my id is {}".format(
            me.full_name,
            me.id,
        )
    )


async def run_polling() -> None:
    async for updates in polling.listen():
        for update in updates:
            lw.add_task(dp.feed(api, update))


lw.add_task(run_polling())
lw.run()



import datetime
from dataclasses import dataclass

from telegrinder import API, Bot, Token, logger, setup_logger
from telegrinder.rules import Argument, Command
from telegrinder.tools.state_mutator import State, StateMutator, mutation

setup_logger()
bot = Bot(API(Token.from_env()))


type Plugged = Stopped


@mutation
async def plug_in() -> Plugged:
    logger.debug("You successfully plugged in your CD player")
    return Stopped()


@dataclass
class Stopped(State):
    @mutation
    def play(self, song: str, offset: datetime.timedelta = datetime.timedelta(0)) -> "Playing":
        return Playing(song, offset, datetime.datetime.now())


@dataclass
class Playing(State):
    song: str
    offset: datetime.timedelta
    started_at: datetime.datetime

    @mutation
    def stop(self) -> "Stopped":
        return Stopped()

    @mutation
    def pause(self) -> "Paused":
        offset = datetime.datetime.now() - self.started_at
        return Paused(
            self.song,
            offset=(offset - self.offset) if self.offset <= datetime.timedelta(0) else offset + self.offset,
            stopped_at=datetime.datetime.now(),
        )


@dataclass
class Paused(State):
    song: str
    offset: datetime.timedelta
    stopped_at: datetime.datetime

    @mutation
    def stop(self) -> "Stopped":
        return Stopped()

    @mutation
    def play(self) -> "Playing":
        return Playing(
            self.song,
            offset=self.offset,
            started_at=datetime.datetime.now(),
        )


@bot.on.message(Command("plug_in"))
async def plug_in_handler(mutator: StateMutator) -> str:
    await plug_in(mutator)
    return "Yes great, choose song with /play <song_name>"


@bot.on.message(Command("play", Argument("song_name", optional=True)))
async def play_song_handler(
    state: Stopped | Paused | Playing,
    song_name: str | None = None,
) -> str:
    match state:
        case Stopped():
            if song_name is None:
                return "You should choose song name"

            await state.play(song_name)
            return f"Started playing {song_name}"
        case Paused() | Playing():
            if song_name is not None:
                previous_song = state.song
                state = await state.stop()
                await state.play(song_name)
                return f"Stopped paused {previous_song}, started playing {song_name}"

            if isinstance(state, Playing):
                return "💽🎵🎼🎶"

            await state.play()
            return f"Continued playing {state.song}"


@bot.on.message(Command("stop"))
async def stop_handler(state: Playing | Paused) -> str:
    match state:
        case Playing():
            await state.stop()
            return f"Stopped playing {state.song} at {datetime.datetime.now() - state.started_at + state.offset}"
        case Paused():
            await state.stop()
            return f"Stopped playing {state.song} at {state.offset}"


@bot.on.message(Command("pause"))
async def pause_handler(state: Playing) -> str:
    new_state = await state.pause()
    return f"Paused {state.song} at {new_state.offset}"


bot.run_forever()



import typing

from nodnod.error import NodeError
from nodnod.interface.scalar import scalar_node

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute


@scalar_node
class CallbackQueryData:
    @classmethod
    def __compose__(cls, callback_query: CallbackQueryCute) -> str:
        return callback_query.data.expect(NodeError("Cannot complete decode callback query data."))


@scalar_node
class CallbackQueryDataJson:
    @classmethod
    def __compose__(cls, callback_query: CallbackQueryCute) -> dict[str, typing.Any]:
        return callback_query.decode_data().expect(
            NodeError("Cannot complete decode callback query data."),
        )


__all__ = ("CallbackQueryData", "CallbackQueryDataJson")



from .fixtures.api_instance import api_instance
from .fixtures.callback_query_update import callback_query_update
from .fixtures.context import callback_query_context, context_factory, message_context
from .fixtures.message_update import message_update
from .fixtures.node_scope import callback_query_node_scope, message_node_scope, node_scope_factory

__all__ = (
    "api_instance",
    "callback_query_context",
    "callback_query_node_scope",
    "callback_query_update",
    "context_factory",
    "message_context",
    "message_node_scope",
    "message_update",
    "node_scope_factory",
)



import typing

from telegrinder.bot.cute_types.inline_query import InlineQueryCute
from telegrinder.bot.dispatch.return_manager.abc import BaseReturnManager, register_manager


class InlineQueryReturnManager(BaseReturnManager):
    @register_manager(dict)
    @staticmethod
    async def dict_manager(handler_response: dict[str, typing.Any], event: InlineQueryCute) -> None:
        await event.answer(**handler_response)


__all__ = ("InlineQueryReturnManager",)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import BusinessMessagesDeleted


class BusinessMessagesDeletedCute(BaseCute[BusinessMessagesDeleted], BusinessMessagesDeleted, kw_only=True):
    pass


__all__ = ("BusinessMessagesDeletedCute",)



from telegrinder import API, Checkbox, Message, Telegrinder, Token, WaiterMachine
from telegrinder.bot.dispatch.waiter_machine.hasher.callback import CALLBACK_QUERY_FOR_MESSAGE
from telegrinder.modules import setup_logger
from telegrinder.rules import Text

setup_logger()

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine()


@bot.on.message(Text("/checkbox"))
async def action(m: Message) -> None:
    picked, m_id = await (
        Checkbox(wm, m.chat.id, "Check your checkbox", cancel_text="Cancel", max_in_row=2)
        .add_option("apple", "Apple", "Apple 🍏")
        .add_option("banana", "Banana", "Banana 🍌", is_picked=True)
        .add_option("pear", "Pear", "Pear 🍐")
        .wait(CALLBACK_QUERY_FOR_MESSAGE, bot.on.callback_query, m.ctx_api)
    )
    await m.edit(
        text="You picked: {}".format(", ".join(c for c in picked if picked[c])),
        chat_id=m.chat.id,
        message_id=m_id,
    )


bot.run_forever()



from telegrinder.tools.serialization.abc import ABCDataSerializer
from telegrinder.tools.serialization.json_ser import JSONSerializer
from telegrinder.tools.serialization.msgpack_ser import MsgPackSerializer

__all__ = ("ABCDataSerializer", "JSONSerializer", "MsgPackSerializer")



import dataclasses
import typing
from abc import ABC, abstractmethod
from datetime import timedelta
from http import HTTPStatus

from telegrinder.client.form_data import MultipartBuilderProto, encode_form_data

if typing.TYPE_CHECKING:
    import datetime

type Data = typing.Any
type Files = dict[str, tuple[str, typing.Any]]
type Timeout = int | float | datetime.timedelta


@dataclasses.dataclass(frozen=True, slots=True)
class Response[T = typing.Any]:
    response: T
    content: bytes
    status: HTTPStatus


class ABCClient(ABC):
    CONNECTION_TIMEOUT_ERRORS: tuple[type[BaseException], ...] = ()
    CLIENT_CONNECTION_ERRORS: tuple[type[BaseException], ...] = ()

    @abstractmethod
    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        pass

    @property
    @abstractmethod
    def timeout(self) -> datetime.timedelta:
        pass

    @abstractmethod
    async def request(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        timeout: int | float | timedelta | None = None,
        **kwargs: typing.Any,
    ) -> Response:
        pass

    @abstractmethod
    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        pass

    @abstractmethod
    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self, **kwargs: typing.Any) -> None:
        pass

    @classmethod
    @abstractmethod
    def multipart_form_builder(cls) -> MultipartBuilderProto:
        pass

    @classmethod
    def get_form(
        cls,
        *,
        data: dict[str, typing.Any] | None = None,
        files: Files | None = None,
    ) -> typing.Any:
        builder = cls.multipart_form_builder()

        if not data and not files:
            return builder.build()

        data = data or {}
        files = files or {}

        for k, v in encode_form_data(data, files).items():
            builder.add_field(k, v)

        for n, (filename, content) in files.items():
            builder.add_field(n, content, filename=filename)

        return builder.build()

    async def __aenter__(self) -> typing.Self:
        return self

    async def __aexit__(
        self,
        exc_type: typing.Any,
        exc_val: typing.Any,
        exc_tb: typing.Any,
    ) -> None:
        await self.close()


__all__ = ("ABCClient", "Response")



"""`Node` module which implements [nodnod](https://github.com/timoniq/nodnod)
integration for convenient dependency injection via `nodes`.


```python
from telegrinder.node import scalar_node, global_node, per_call

@global_node
@scalar_node
class Hello:
    @classmethod
    def __compose__(cls) -> str:
        return "Hello"

@global_node
@scalar_node
class World:
    @classmethod
    def __compose__(cls) -> str:
        return "World"

@per_call
@scalar_node
class Greeter:
    @classmethod
    def __compose__(cls, hello: Hello, world: World) -> str:
        return f"{hello}, {world}!"

@bot.on.message()
async def hello_world(greeter: Greeter) -> str:
    return greeter
```
"""

# pyright: reportUnusedImport=false, reportUnsupportedDunderAll=false

from nodnod import (
    DataNode,
    Injection,
    Node,
    NodeConstructor,
    ResultNode,
    Scalar,
    case,
    generic_node,
    polymorphic,
    scalar_node,
)

from telegrinder.node.compose import Composable, compose, create_composable, run_agent
from telegrinder.node.nodes import *
from telegrinder.node.nodes import __all__ as nodes_all
from telegrinder.node.scope import GLOBAL, PER_CALL, PER_EVENT, NodeScope, global_node, per_call, per_event
from telegrinder.node.utils import as_node, is_node

__all__ = nodes_all + (
    "GLOBAL",
    "PER_CALL",
    "PER_EVENT",
    "Composable",
    "DataNode",
    "NodeConstructor",
    "NodeScope",
    "Node",
    "ResultNode",
    "Injection",
    "Scalar",
    "as_node",
    "compose",
    "create_composable",
    "run_agent",
    "global_node",
    "is_node",
    "case",
    "generic_node",
    "polymorphic",
    "scalar_node",
    "per_call",
    "per_event",
)



from telegrinder import API, Message, Telegrinder, Token, setup_logger
from telegrinder.rules import Text
from telegrinder.tools.formatting.html import HTML, bold, italic, mention

setup_logger()
api = API(token=Token.from_env())
api.default_params["parse_mode"] = HTML.PARSE_MODE

bot = Telegrinder(api)


@bot.on.message(Text("/formatting"))
async def formatting(m: Message):
    await m.answer(bold(italic("bold italic text!")))
    await m.answer(HTML << "this is " << mention(m.from_user.first_name, user_id=m.from_user.id))


bot.run_forever()



import typing

from kungfu.library import Error, Ok, Result, unwrapping
from nodnod.error import NodeError

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.rules.abc import ABCRule, Always, AndRule
from telegrinder.modules import logger
from telegrinder.node.compose import compose
from telegrinder.tools import fullname
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

type Handler = typing.Callable[..., typing.Any]
type ActionFunction = typing.Callable[..., ActionFunctionResult]
type ActionFunctionResult = typing.Union[
    typing.AsyncGenerator[typing.Any, typing.Any],
    typing.Awaitable[typing.Any],
    typing.Generator[typing.Any, typing.Any, typing.Any],
    typing.Any,
]


@unwrapping
async def run_action_function[T: Handler](
    func_handler: FuncHandler[T],
    function: ActionFunction,
    api: API,
    update: Update,
    context: Context,
    agent: type[Agent] | None = None,
) -> Result[typing.Any, str]:
    async with compose(
        function,
        context,
        agent_cls=agent,
    ) as result:
        match result:
            case Ok():
                return await func_handler.run(api, update, context)
            case Error(error):
                return Error(
                    "{}\n".format(
                        NodeError(f"* failed to compose action function `{fullname(function)}`", from_error=error),
                    ),
                )


def action(function: ActionFunction, agent: type[Agent] | None = None) -> Action:
    return Action(function, agent=agent)


class Action:
    _on: ABCRule

    def __init__(self, function: ActionFunction, agent: type[Agent] | None = None) -> None:
        self.function = function
        self.agent = agent
        self._on = Always()

    def on(self, *rules: ABCRule) -> typing.Self:
        self._on &= AndRule(*rules)
        return self

    def __call__[T: Handler](self, handler: T) -> T:
        func_handler = FuncHandler(function=handler, agent=self.agent)

        async def action_wrapper(api: API, update: Update, context: Context) -> typing.Any:
            if not await check_rule(self._on, context):
                await logger.adebug("Action rule `{!r}` failed.", self._on)
                result = await func_handler.run(api, update, context)
            else:
                result = await run_action_function(
                    func_handler,
                    self.function,
                    api,
                    update,
                    context,
                    agent=self.agent,
                )

            match result:
                case Ok(value):
                    return value
                case Error(error):
                    await logger.adebug(error)
                    return None

        action_wrapper.__name__ = f"<action for {handler.__name__}>"
        action_wrapper.__qualname__ = f"<action for {handler.__qualname__}>"
        return action_wrapper  # type: ignore


__all__ = ("Action", "action")



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import MessageReactionCountUpdated


class MessageReactionCountUpdatedCute(BaseCute[MessageReactionCountUpdated], MessageReactionCountUpdated, kw_only=True):
    pass


__all__ = ("MessageReactionCountUpdatedCute",)



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import setup_logger
from telegrinder.node import Error
from telegrinder.rules import IsUser, Text

setup_logger()

api = API(token=Token.from_env())
bot = Telegrinder(api)


@bot.on.message(Text("oops"))
async def oops_handler(m: Message):
    await m.answer("Oh no")
    raise RuntimeError("Wow")


@bot.on.message(Text("woops"))
def woops_handler():
    raise ValueError("Wow oopsii!")


@bot.on.event_error(IsUser())
async def error_handler(err: Error[RuntimeError, ValueError], m: Message):
    await m.answer(f"okay..( Something happened: {err.exception}")


bot.run_forever()



import typing

from nodnod.error import NodeError
from nodnod.interface.node_constructor import NodeConstructor
from nodnod.interface.scalar import scalar_node

from telegrinder.bot.cute_types.message import MessageCute


@scalar_node
class Caption:
    @classmethod
    def __compose__(cls, message: MessageCute) -> str:
        return message.caption.expect(NodeError("Message has no caption."))


@scalar_node
class HTMLCaption:
    @classmethod
    def __compose__(cls, message: MessageCute) -> str:
        return message.html_caption.expect(NodeError("Message has no HTML caption."))


@scalar_node
class Text:
    @classmethod
    def __compose__(cls, message: MessageCute) -> str:
        return message.text.expect(NodeError("Message has no text."))


@scalar_node
class HTMLText:
    @classmethod
    def __compose__(cls, message: MessageCute) -> str:
        return message.html_text.expect(NodeError("Message has no HTML text."))


@scalar_node
class TextInteger:
    @classmethod
    def __compose__(cls, text: Text | Caption) -> int:
        if not text.isdigit():
            raise NodeError("Text is not digit.")
        return int(text)


if typing.TYPE_CHECKING:
    from typing import Literal as TextLiteral

else:

    class TextLiteral(NodeConstructor):
        def __init__(self, *texts: str) -> None:
            self.texts = texts

        def __compose__(self, text: Text | Caption) -> str:
            if text in self.texts:
                return text
            raise NodeError("Text mismatched literal.")


__all__ = ("Caption", "HTMLCaption", "HTMLText", "Text", "TextInteger", "TextLiteral")



import dataclasses
import typing

from kungfu.library.monad.option import Nothing, Option, Some
from nodnod.error import NodeError
from nodnod.interface.polymorphic import case, polymorphic
from nodnod.interface.scalar import scalar_node

from telegrinder.api.api import API
from telegrinder.bot.cute_types import (
    CallbackQueryCute,
    ChatJoinRequestCute,
    ChatMemberUpdatedCute,
    ChosenInlineResultCute,
    InlineQueryCute,
    MessageCute,
    MessageReactionUpdatedCute,
    PaidMediaPurchasedCute,
    PollAnswerCute,
    PreCheckoutQueryCute,
    ShippingQueryCute,
)
from telegrinder.types.objects import Chat, Message, User


@scalar_node
@polymorphic["Source"]
@dataclasses.dataclass(kw_only=True)
class Source:
    api: API
    from_user: User
    chat: Option[Chat] = dataclasses.field(default_factory=Nothing)
    thread_id: Option[int] = dataclasses.field(default_factory=Nothing)

    @case
    def compose_message(cls, message: MessageCute) -> typing.Self:
        return cls(
            api=message.api,
            from_user=message.from_.expect(NodeError("Message is from a channel.")),
            chat=Some(message.chat),
            thread_id=message.message_thread_id,
        )

    @case
    def compose_callback_query(cls, callback_query: CallbackQueryCute) -> typing.Self:
        return cls(
            api=callback_query.api,
            from_user=callback_query.from_user,
            chat=callback_query.chat,
            thread_id=callback_query.message_thread_id,
        )

    @case
    def compose_inline_query(cls, inline_query: InlineQueryCute) -> typing.Self:
        return cls(
            api=inline_query.api,
            from_user=inline_query.from_user,
        )

    @case
    def compose_chosen_inline_result(cls, chosen_inline_result: ChosenInlineResultCute) -> typing.Self:
        return cls(
            api=chosen_inline_result.api,
            from_user=chosen_inline_result.from_user,
        )

    @case
    def compose_chat_member_updated(cls, chat_member_updated: ChatMemberUpdatedCute) -> typing.Self:
        return cls(
            api=chat_member_updated.api,
            from_user=chat_member_updated.from_user,
            chat=Some(chat_member_updated.chat),
        )

    @case
    def compose_chat_join_request(cls, chat_join_request: ChatJoinRequestCute) -> typing.Self:
        return cls(
            api=chat_join_request.api,
            from_user=chat_join_request.from_user,
            chat=Some(chat_join_request.chat),
        )

    @case
    def compose_pre_checkout_query(cls, pre_checkout_query: PreCheckoutQueryCute) -> typing.Self:
        return cls(
            api=pre_checkout_query.api,
            from_user=pre_checkout_query.from_user,
        )

    @case
    def compose_message_reaction_updated(cls, message_reaction_updated: MessageReactionUpdatedCute) -> typing.Self:
        return cls(
            api=message_reaction_updated.api,
            from_user=message_reaction_updated.user.expect(NodeError("Message reaction is from an anonymous user.")),
            chat=Some(message_reaction_updated.chat),
        )

    @case
    def compose_paid_media_purchased(cls, paid_media_purchased: PaidMediaPurchasedCute) -> typing.Self:
        return cls(
            api=paid_media_purchased.api,
            from_user=paid_media_purchased.from_user,
        )

    @case
    def compose_poll_answer(cls, poll_answer: PollAnswerCute) -> typing.Self:
        return cls(
            api=poll_answer.api,
            from_user=poll_answer.user.expect(NodeError("Poll answer is from an anonymous user.")),
            chat=poll_answer.voter_chat,
        )

    @case
    def compose_shipping_query(cls, shipping_query: ShippingQueryCute) -> typing.Self:
        return cls(
            api=shipping_query.api,
            from_user=shipping_query.from_user,
        )

    async def send(self, text: str, **kwargs: typing.Any) -> Message:
        result = await self.api.send_message(
            chat_id=self.chat.map_or(self.from_user.id, lambda chat: chat.id).unwrap(),
            message_thread_id=self.thread_id.unwrap_or_none(),
            text=text,
            **kwargs,
        )
        return result.unwrap()


@scalar_node
class ChatSource:
    @classmethod
    def __compose__(cls, source: Source) -> Chat:
        return source.chat.expect(NodeError("Source has no chat."))


@scalar_node
class UserSource:
    @classmethod
    def __compose__(cls, source: Source) -> User:
        return source.from_user


@scalar_node
class ChatId:
    @classmethod
    def __compose__(cls, chat: ChatSource) -> int:
        return chat.id


@scalar_node
class UserId:
    @classmethod
    def __compose__(cls, user: UserSource) -> int:
        return user.id


@scalar_node
class Locale:
    @classmethod
    def __compose__(cls, user: UserSource) -> str:
        return user.language_code.expect(NodeError("User has no language code."))


__all__ = (
    "ChatId",
    "ChatSource",
    "Locale",
    "Source",
    "UserId",
    "UserSource",
)



import typing
from abc import ABC, abstractmethod

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.types.objects import Update


class ABCRouter(ABC):
    @abstractmethod
    async def route(self, api: API, update: Update, context: Context) -> typing.Any:
        pass


__all__ = ("ABCRouter",)



from __future__ import annotations

import typing
from functools import cached_property

from kungfu.library import Some, Sum
from kungfu.library.misc import from_optional
from nodnod.error import NodeError

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.model import Model
from telegrinder.msgspec_utils import Option, encoder, get_class_annotations, struct_asdict
from telegrinder.tools.fullname import fullname
from telegrinder.tools.magic.shortcut import shortcut
from telegrinder.types.objects import Update

BOUND_API_KEY: typing.Final = "bound_api"
RAW_UPDATE_BIND_KEY: typing.Final = "raw_update_bind"


def compose_method_params[Cute: BaseCute](
    params: dict[str, typing.Any],
    update: Cute,
    *,
    default_params: set[str | tuple[str, str]] | None = None,
    validators: dict[str, typing.Callable[[Cute], bool]] | None = None,
) -> dict[str, typing.Any]:
    default_params = default_params or set()
    validators = validators or {}

    for param in default_params:
        param_name = param if isinstance(param, str) else param[0]
        if param_name not in params:
            if param_name in validators and not validators[param_name](update):
                continue
            params[param_name] = getattr(update, param if isinstance(param, str) else param[1], None)

    return params


def get_cute_from_generic(generic_args: tuple[typing.Any, ...], /) -> typing.Any:
    for arg in generic_args:
        orig_arg = typing.get_origin(arg) or arg

        if not isinstance(orig_arg, type):
            continue

        if orig_arg in (Sum, Some, Option):
            return get_cute_from_generic(typing.get_args(arg))

        if issubclass(orig_arg, BaseCute):
            return arg

    return None


def get_cute_annotations(annotations: dict[str, typing.Any], /) -> dict[str, type[BaseCute]]:
    cute_annotations = {}

    for key, hint in annotations.items():
        if not isinstance(hint, type) and (cute := get_cute_from_generic(typing.get_args(hint))) is not None:
            cute_annotations[key] = cute

        elif isinstance(hint, type) and issubclass(hint, BaseCute):
            cute_annotations[key] = hint

    return cute_annotations


def maybe_wrapped(value: typing.Any, /) -> typing.Any:
    wrapped_value = None

    while isinstance(value, Sum | Some):
        wrapped_value = value

        if isinstance(value, Sum):
            value = value.v

        if isinstance(value, Some):
            value = value.value

    return wrapped_value if wrapped_value is not None else value


def wrap_value(value: typing.Any, type_: typing.Any, /) -> typing.Any:
    args = [type_]
    types = []

    while args:
        arg = args.pop(0)
        origin_arg = typing.get_origin(arg) or arg

        if issubclass(origin_arg, Sum | Option):
            args.extend(typing.get_args(arg))
            types.append(Some if issubclass(origin_arg, Option) else arg)  # type: ignore

    result = value

    for t in types[::-1]:
        result = t(result)

    return result


def to_cute(
    cute_cls: type[BaseCute],
    field: str,
    value: typing.Any,
    bound_api: API,
) -> typing.Any:
    maybe_wrapped_value = maybe_wrapped(value)
    is_wrapped_value = isinstance(maybe_wrapped_value, Sum | Some)
    cute_annotations = cute_cls.__cute_annotations__

    if cute_annotations is None:
        cute_annotations = get_cute_annotations(get_class_annotations(cute_cls))

    cute = cute_annotations[field].from_update(
        maybe_wrapped_value._value if is_wrapped_value else maybe_wrapped_value,
        bound_api=bound_api,
    )
    return wrap_value(cute, cute_cls.__annotations__[field]) if is_wrapped_value else cute


class BaseCute[T: Model = typing.Any](Model):
    def __init_subclass__(cls, *args: typing.Any, **kwargs: typing.Any) -> None:
        cls.__is_resolved_annotations__ = False
        cls.__cute_annotations__ = None
        super().__init_subclass__(*args, **kwargs)

    @classmethod
    def __compose__(cls, update: Update, context: Context) -> typing.Any:
        update_cute = context.update_cute

        if type(update_cute) is cls:
            return update_cute

        if type(update_cute.incoming_update) is cls:
            return update_cute.incoming_update

        raise NodeError(f"Incoming update is not `{fullname(cls)}`.")

    @classmethod
    def from_update(cls, update: Update, bound_api: API) -> typing.Self:
        if not cls.__is_resolved_annotations__:
            cls.__is_resolved_annotations__ = True
            cls.__annotations__ = get_class_annotations(cls)

        if cls.__cute_annotations__ is None:
            cls.__cute_annotations__ = get_cute_annotations(cls.__annotations__)

        cute = cls(
            **{
                field: to_cute(cls, field, value, bound_api) if field in cls.__cute_annotations__ else value
                for field, value in update.to_dict().items()
            },
        )._set_bound_api(api=bound_api)
        return cute.bind_raw_update(update) if isinstance(update, Update) else cute

    @cached_property
    def bound_api(self) -> API: ...

    @property
    def api(self) -> API:
        return self.bound_api

    @property
    def ctx_api(self) -> API:
        return self.bound_api

    @property
    def raw_update(self) -> Update:
        return self.get_raw_update().expect(
            AttributeError(f"Cute model `{type(self).__name__}` has no bound `Update` object."),
        )

    def get_raw_update(self) -> Option[Update]:
        return from_optional(self.__dict__.get(RAW_UPDATE_BIND_KEY))

    def bind_raw_update(self, raw_update: Update, /) -> typing.Self:
        cuties = [self]

        if isinstance(self, Update) and isinstance(cute := self.incoming_update, BaseCute):
            cuties.append(cute)

        for cute in cuties:
            cute.__dict__[RAW_UPDATE_BIND_KEY] = raw_update  # type: ignore

        return self

    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
        full: bool = False,
    ) -> dict[str, typing.Any]:
        return self._to_dict(
            "model_as_dict" if not full else "model_as_full_dict",
            exclude_fields=exclude_fields or set(),
            full=full,
        )

    def to_full_dict(self, *, exclude_fields: set[str] | None = None) -> dict[str, typing.Any]:
        return self.to_dict(exclude_fields=exclude_fields, full=True)

    def _set_bound_api(self, api: API) -> typing.Self:
        self.__dict__[BOUND_API_KEY] = api  # type: ignore
        return self

    def _to_dict(self, dct_name: str, exclude_fields: set[str], full: bool) -> dict[str, typing.Any]:
        if dct_name not in self.__dict__:
            dct = struct_asdict(self)
            self.__dict__[dct_name] = dct if not full else encoder.to_builtins(dct, order="deterministic")  # type: ignore

        if not exclude_fields:
            return self.__dict__[dct_name]

        return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}


class BaseShortcuts[T: BaseCute = typing.Any]:
    @cached_property
    def cute(self) -> typing.Self:
        return self


__all__ = ("BaseCute", "BaseShortcuts", "compose_method_params", "shortcut")



import sys
import typing

from nodnod.interface.is_node import is_node
from nodnod.node import Node

from telegrinder.tools.fullname import fullname


def as_node(obj: typing.Any, /) -> type[Node[typing.Any, typing.Any]]:
    if not is_node(obj):
        raise TypeError(f"Object `{fullname(obj)}` is not a node.")
    return obj


def get_locals_from_function(func: typing.Callable[..., typing.Any], /) -> typing.Mapping[str, typing.Any] | None:
    bound_class = func.__self__ if hasattr(func, "__self__") else None
    bound_class = func.__objclass__ if bound_class is None and hasattr(func, "__objclass__") else None

    if bound_class is not None:
        return (bound_class if isinstance(bound_class, type) else type(bound_class)).__dict__

    return None


def get_globals_from_function(func: typing.Callable[..., typing.Any], /) -> dict[str, typing.Any]:
    if hasattr(func, "__globals__"):
        return getattr(func, "__globals__")

    module = getattr(func, "__module__", type(func).__module__)

    if module in sys.modules:
        return vars(sys.modules[module])

    return {}


__all__ = ("as_node", "get_globals_from_function", "get_locals_from_function", "is_node")



import typing

import msgspec
from kungfu.library.monad.option import Nothing

UNSET: typing.Final = typing.cast("typing.Any", msgspec.UNSET)

def is_none(obj: typing.Any, /) -> typing.TypeIs[Nothing | None]: ...

@typing.overload
def field() -> typing.Any: ...
@typing.overload
def field(*, name: str | None = ...) -> typing.Any: ...
@typing.overload
def field(*, default: typing.Any, name: str | None = ...) -> typing.Any: ...
@typing.overload
def field(*, default_factory: typing.Callable[[], typing.Any], name: str | None = None) -> typing.Any: ...
@typing.overload
def field(*, converter: typing.Callable[[typing.Any], typing.Any], name: str | None = ...) -> typing.Any: ...
@typing.overload
def field(*, default: typing.Any, converter: typing.Callable[[typing.Any], typing.Any], name: str | None = ...) -> typing.Any: ...
@typing.overload
def field(
    *,
    default_factory: typing.Callable[[], typing.Any],
    converter: typing.Callable[[typing.Any], typing.Any],
    name: str | None = None,
) -> typing.Any: ...

class From[T]:
    def __new__(cls, _: T, /) -> typing.Any: ...

@typing.dataclass_transform(field_specifiers=(field,))
class Model(msgspec.Struct, dict=True):
    @classmethod
    def get_fields(cls) -> typing.Mapping[str, msgspec.inspect.Field]: ...
    @classmethod
    def get_optional_fields(cls) -> typing.Iterable[str]: ...
    @classmethod
    def from_data[**P, T](cls: typing.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T: ...
    @classmethod
    def from_dict(cls, obj: dict[str, typing.Any], /) -> typing.Self: ...
    @classmethod
    def from_raw(cls, raw: str | bytes, /) -> typing.Self: ...
    def to_raw(self) -> str: ...
    def to_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]: ...
    def to_full_dict(
        self,
        *,
        exclude_fields: set[str] | None = None,
    ) -> dict[str, typing.Any]: ...

__all__ = ("UNSET", "From", "Model", "field", "is_none")



from nodnod.error import NodeError
from nodnod.interface.scalar import scalar_node

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.types.objects import MessageEntity


@scalar_node
class MessageEntities:
    @classmethod
    def __compose__(cls, message: MessageCute) -> list[MessageEntity]:
        return message.entities.expect(NodeError("Message has no entities."))


__all__ = ("MessageEntities",)



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.types.objects import ChosenInlineResult, User


class ChosenInlineResultCute(BaseCute[ChosenInlineResult], ChosenInlineResult, kw_only=True):
    @property
    def from_user(self) -> User:
        return self.from_


__all__ = ("ChosenInlineResultCute",)



from telegrinder.bot.rules.abc import ABCRule
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update


class IsUpdateType(ABCRule):
    def __init__(self, update_type: UpdateType, /) -> None:
        self.update_type = update_type

    def check(self, update: Update) -> bool:
        return update.update_type == self.update_type


__all__ = ("IsUpdateType",)



import typing

_NOVALUE: typing.Any = object()


def private_func_name(func: typing.Callable[..., typing.Any], /) -> str:
    return "__" + func.__name__


class additional_property[T, R]:  # noqa: N801
    def __init__(self, func: typing.Callable[[T], R], /) -> None:
        self.func = func
        self.func_name = private_func_name(func)

    def __set__(self, instance: T, value: typing.Any, /) -> None:
        instance.__dict__[self.func_name] = value

    def __get__(self, instance: T | None, owner: type[T], /) -> typing.Any:
        if instance is None:
            raise AttributeError("Cannot access additional property from class.")

        value = instance.__dict__.get(self.func_name, _NOVALUE)
        if value is _NOVALUE:
            value = self.func(instance)
            instance.__dict__[self.func_name] = value

        return value

    def __delete__(self, instance: T, /) -> None:
        instance.__dict__.pop(self.func_name, None)


class class_property[T]:  # noqa: N801
    def __init__(self, func: typing.Callable[..., T], /) -> None:
        self.func = func
        self.func_name = private_func_name(func)

    def __get__(self, instance: typing.Any | None, owner: type[typing.Any], /) -> T:
        return self.func(owner)


class class_cached_property[T]:  # noqa: N801
    def __init__(self, func: typing.Callable[..., T], /) -> None:
        self.func = func
        self.func_name = private_func_name(func)

    def __get__(self, instance: typing.Any | None, owner: type[typing.Any], /) -> T:
        value = getattr(owner, self.func_name, _NOVALUE)

        if value is _NOVALUE:
            value = self.func(owner)
            setattr(owner, self.func_name, value)

        return value


__all__ = ("additional_property", "class_cached_property", "class_property")



import typing
from abc import ABC, abstractmethod
from collections import deque
from functools import cached_property

from nodnod.agent.event_loop.agent import EventLoopAgent
from nodnod.interface.node_from_function import create_node_from_function

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.node.compose import create_composable
from telegrinder.node.utils import get_globals_from_function, get_locals_from_function
from telegrinder.tools.fullname import fullname

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

    from telegrinder.node.compose import Composable

type CheckResult = bool | typing.Awaitable[bool]
type Node = typing.Any


class ABCRule(ABC):
    required_nodes: typing.Mapping[str, Node] | None = None
    agent_cls: type[Agent] = EventLoopAgent
    requires: deque[ABCRule] = deque()

    @abstractmethod
    def check(self, *args: typing.Any, **kwargs: typing.Any) -> CheckResult:
        pass

    def __init_subclass__(
        cls,
        *,
        requires: typing.Iterable[ABCRule] | None = None,
    ) -> None:
        requirements: list[ABCRule] = []

        for base in cls.__mro__:
            if issubclass(base, ABCRule) and base != cls:
                requirements.extend(base.requires or ())

        requirements.extend(requires or ())
        cls.requires = deque(dict.fromkeys(requirements))

    def __and__(self, other: object, /) -> AndRule:
        if not isinstance(other, ABCRule):
            return NotImplemented
        return AndRule(self, other)

    def __iadd__(self, other: object, /) -> AndRule:
        return self.__and__(other)

    def __or__(self, other: object, /) -> OrRule:
        if not isinstance(other, ABCRule):
            return NotImplemented
        return OrRule(self, other)

    def __ior__(self, other: object, /) -> OrRule:
        return self.__or__(other)

    def __invert__(self) -> NotRule:
        return NotRule(self)

    def __repr__(self) -> str:
        return "<{}{}>".format(
            fullname(self),
            "" if not self.requires else ", requires={!r}".format(self.requires),
        )

    def as_optional(self) -> ABCRule:
        return self | Always()

    def should_fail(self) -> ABCRule:
        return self & Never()

    @cached_property
    def composable(self) -> Composable:
        node = create_node_from_function(
            self.check,
            dependencies=self.required_nodes,
            forward_refs=get_globals_from_function(self.check),
            namespace=get_locals_from_function(self.check),
        )
        return create_composable(node, agent_cls=self.agent_cls)


class AndRule(ABCRule):
    def __init__(self, *rules: ABCRule) -> None:
        self.rules = rules

    async def check(self, context: Context) -> bool:
        for rule in self.rules:
            if not await check_rule(rule, context):
                return False

        return True


class OrRule(ABCRule):
    def __init__(self, *rules: ABCRule) -> None:
        self.rules = rules

    async def check(self, context: Context) -> bool:
        for rule in self.rules:
            if await check_rule(rule, context):
                return True

        return False


class NotRule(ABCRule):
    def __init__(self, rule: ABCRule) -> None:
        self.rule = rule

    async def check(self, context: Context) -> bool:
        return not await check_rule(self.rule, context)


class Never(ABCRule):
    """Neutral element for `|` (OrRule)."""

    def check(self) -> typing.Literal[False]:
        return False


class Always(ABCRule):
    """Neutral element for `&` (AndRule)."""

    def check(self) -> typing.Literal[True]:
        return True


__all__ = (
    "ABCRule",
    "Always",
    "AndRule",
    "CheckResult",
    "Never",
    "NotRule",
    "OrRule",
    "check_rule",
)



import enum
import html
import types
import typing
from string.templatelib import Template

from telegrinder.tools.formatting.deep_links.links import tg_mention_link
from telegrinder.tools.parse_mode import ParseMode

type FormatString = object | Template
type Formatter = typing.Callable[[FormatString], TagFormat]

TAG_FORMAT: typing.Final = "<{tag}{data}>{content}</{tag}>"
UNION_SPECIFIERS_SEPARATOR: typing.Final = "+"


def blockquote(s: FormatString, /, *, expandable: bool = False) -> TagFormat:
    return TagFormat(s, tag=Tag.BLOCK_QUOTE, **dict(expandable=None) if expandable is True else dict())


def bold(s: FormatString, /) -> TagFormat:
    return TagFormat(s, tag=Tag.BOLD)


def code_inline(s: FormatString, /) -> TagFormat:
    return TagFormat(s, tag=Tag.CODE)


def expandable_blockquote(s: FormatString, /) -> TagFormat:
    return blockquote(s, expandable=True)


monospace = code_inline


def italic(s: FormatString, /) -> TagFormat:
    return TagFormat(s, tag=Tag.ITALIC)


def link(href: FormatString, /, *, text: str | None = None) -> TagFormat:
    return TagFormat(
        text or href,
        tag=Tag.HYPERLINK,
        href=f'"{href}"',
    )


def pre_code(s: FormatString, /, *, lang: str | None = None) -> TagFormat:
    if not lang:
        return TagFormat(s, tag=Tag.PRE)
    return pre_code(TagFormat(s, tag=Tag.CODE, **{"class": f"language-{lang}"}))


def spoiler(s: FormatString, /) -> TagFormat:
    return TagFormat(s, tag=Tag.SPOILER)


def strike(s: FormatString, /) -> TagFormat:
    return TagFormat(s, tag=Tag.STRIKE)


def mention(s: FormatString, /, *, user_id: int) -> TagFormat:
    text = escape(s) if isinstance(s, Template) else str(s)
    return link(tg_mention_link(user_id=user_id), text=text)


def tg_emoji(s: FormatString, /, *, emoji_id: int) -> TagFormat:
    return TagFormat(s, tag=Tag.EMOJI, emoji_id=emoji_id)


def underline(s: FormatString, /) -> TagFormat:
    return TagFormat(s, tag=Tag.UNDERLINE)


def _apply_formats(value: typing.Any, format_spec: str | None = None) -> str:
    if not format_spec:
        return html.escape(str(value))

    result: FormatString = html.escape(str(value))

    for fmt in (f.strip() for f in format_spec.split(UNION_SPECIFIERS_SEPARATOR)):
        if fmt in FORMATTERS:
            result = FORMATTERS[fmt](result)
        else:
            raise ValueError(f"Unknown format: `{fmt}`")

    return result.formatting() if isinstance(result, TagFormat) else result


def escape(s: FormatString, /) -> str:
    if isinstance(s, Template):
        parts: list[str] = []

        for item in s:
            if isinstance(item, str):
                parts.append(html.escape(item))
            else:
                value = item.value

                if isinstance(value, TagFormat):
                    parts.append(value.formatting())
                elif item.format_spec:
                    parts.append(_apply_formats(value, item.format_spec))
                else:
                    parts.append(html.escape(str(value)))

        return "".join(parts)

    return html.escape(str(s))


class HTMLMeta(type):
    def __lshift__[T](cls: typing.Callable[..., T], other: object, /) -> T:
        if not isinstance(other, str | Template | TagFormat):
            return NotImplemented
        return cls(other)


class Tag(enum.StrEnum):
    HYPERLINK = "a"
    BOLD = "b"
    ITALIC = "i"
    UNDERLINE = "u"
    STRIKE = "s"
    CODE = "code"
    PRE = "pre"
    SPOILER = "tg-spoiler"
    BLOCK_QUOTE = "blockquote"
    EMOJI = "tg-emoji"

    def __str__(self) -> str:
        return self.value


class TagFormat(str):
    __slots__ = ("data", "tag")

    tag: Tag
    data: dict[str, typing.Any]

    def __new__(cls, s: FormatString, /, *, tag: Tag, **data: typing.Any) -> typing.Self:
        s = escape(s) if not isinstance(s, TagFormat) else s.formatting()
        obj = super().__new__(cls, s)
        obj.data = data
        obj.tag = tag
        return obj

    def __str__(self) -> str:
        return self.formatting()

    @property
    def content(self) -> str:
        return super().__str__()

    @property
    def tag_data(self) -> str:
        return "".join(f" {k}={v}" if v is not None else f" {k}" for k, v in self.data.items())

    def formatting(self) -> str:
        return TAG_FORMAT.format(
            tag=self.tag,
            data=self.tag_data,
            content=self.content,
        )


class HTML(str, metaclass=HTMLMeta):
    """A class for creating and formatting HTML strings.

    >>> HTML << "Hello, " << bold("World!")
    'Hello, <b>World!</b>'

    >>> HTML << t"Hello, {name:underline}! " << "You are <b>{age}</b> years old."
    'Hello, <u>Arseny</u>! You are <b>20</b> years old.'
    """

    PARSE_MODE: typing.Final = ParseMode.HTML

    __slots__ = ()

    def __new__(cls, s: str | Template | TagFormat, /) -> typing.Self:
        if isinstance(s, Template):
            content = escape(s)
        elif isinstance(s, TagFormat):
            content = s.formatting()
        else:
            content = s

        return super().__new__(cls, content)

    def __lshift__(self, other: object, /) -> typing.Self:
        if isinstance(other, TagFormat):
            addition = other.formatting()
        elif isinstance(other, Template):
            addition = escape(other)
        else:
            addition = str(other)

        return self.__class__.__new__(self.__class__, self + addition)


FORMATTERS: types.MappingProxyType[str, Formatter] = types.MappingProxyType(
    mapping=dict(
        bold=bold,
        b=bold,
        code=code_inline,
        c=code_inline,
        monospace=monospace,
        italic=italic,
        i=italic,
        spoiler=spoiler,
        sp=spoiler,
        strike=strike,
        s=strike,
        underline=underline,
        u=underline,
        blockquote=blockquote,
        expandable_blockquote=expandable_blockquote,
    ),
)


__all__ = (
    "HTML",
    "Tag",
    "TagFormat",
    "blockquote",
    "bold",
    "code_inline",
    "escape",
    "expandable_blockquote",
    "italic",
    "link",
    "mention",
    "monospace",
    "pre_code",
    "spoiler",
    "strike",
    "tg_emoji",
    "underline",
)



import enum

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.node.nodes.text import Text


class EnumTextRule[T: enum.Enum](ABCRule):
    def __init__(self, enum_t: type[T], *, lower_case: bool = True) -> None:
        self.lower_case = lower_case
        self.enum_t = enum_t
        self.enumerations_map = {
            (enumeration.value.lower() if lower_case else enumeration.value): enumeration for enumeration in self.enum_t
        }

    def find(self, s: str) -> T | None:
        s = s.lower() if self.lower_case else s
        return self.enumerations_map.get(s)

    def check(self, text: Text, ctx: Context) -> bool:
        match self.find(text):
            case enum.Enum() as enumeration:
                ctx.enum_text = enumeration
                return True
            case _:
                return False


__all__ = ("EnumTextRule",)



def to_utf16_map(s: str, /) -> list[int]:
    utf16_map = list[int]()
    utf16_pos = 0

    for char in s:
        utf16_map.append(utf16_pos)
        utf16_len = len(char.encode("utf-16-le")) // 2
        utf16_pos += utf16_len

    utf16_map.append(utf16_pos)
    return utf16_map


def utf16_to_py_index(utf16_map: list[int], utf16_index: int, /) -> int:
    for index, u in enumerate(utf16_map):
        if u >= utf16_index:
            return index

    return len(utf16_map) - 1


__all__ = ("to_utf16_map", "utf16_to_py_index")



from telegrinder import API, Dispatch, Telegrinder, Token, WaiterMachine
from telegrinder.modules import setup_logger

setup_logger()

dp = Dispatch()
bot = Telegrinder(API(Token.from_env()), dispatch=dp)
wm = WaiterMachine()



import typing

from nodnod.error import NodeError
from nodnod.interface.node_constructor import NodeConstructor

import telegrinder.types as tg_types
from telegrinder.api.api import API
from telegrinder.node.nodes.attachment import Animation, Audio, Document, Photo, Video, VideoNote, Voice

type Attachment = Animation | Audio | Document | Photo | Video | VideoNote | Voice


class FileIdNode(NodeConstructor):
    def __init__(self, attachment_node: type[Attachment], /) -> None:
        self.__map__ = {Attachment: attachment_node}

    def __compose__(self, attach: Attachment) -> str:
        if isinstance(attach, Photo):
            return attach.sizes[-1].file_id
        return attach.file_id


class FileNode(NodeConstructor):
    def __init__(self, attachment_node: type[Attachment], /) -> None:
        self.__map__ = {str: FileIdNode[attachment_node]}

    async def __compose__(self, api: API, file_id: str) -> tg_types.File:
        return (await api.get_file(file_id=file_id)).expect(NodeError("File can't be downloaded."))


if typing.TYPE_CHECKING:
    type FileId[T: Attachment] = str
    type File[T: Attachment] = tg_types.File
else:
    FileId = FileIdNode
    File = FileNode


__all__ = ("File", "FileId")



import types
import typing


def _get_types(x: typing.Any, /) -> type[typing.Any] | tuple[typing.Any, ...]:
    while True:
        if isinstance(x, types.UnionType | typing._UnionGenericAlias):  # type: ignore
            return tuple(_get_types(x) for x in typing.get_args(x))

        if isinstance(x, typing.TypeAliasType):
            x = x.__value__

        if isinstance(x, types.GenericAlias | typing._GenericAlias):  # type: ignore
            x = typing.get_origin(x)

        if isinstance(x, type):
            return x


__all__ = ("_get_types",)



import pathlib
import random
from datetime import timedelta
from functools import partial

from telegrinder import (
    API,
    CALLBACK_QUERY_FOR_MESSAGE,
    MESSAGE_FROM_USER_IN_CHAT,
    Context,
    InlineButton,
    InlineKeyboard,
    Message,
    Telegrinder,
    Token,
    configure_dotenv,
    setup_logger,
)
from telegrinder.bot import WaiterMachine
from telegrinder.bot.dispatch.handler import MessageReplyHandler
from telegrinder.bot.dispatch.middleware import ABCMiddleware
from telegrinder.node import Me, UserId
from telegrinder.rules import (
    CallbackDataEq,
    FuzzyText,
    HasText,
    IsUpdateType,
    IsUser,
    Markup,
    Text,
)
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import InputFile

configure_dotenv(load_file=True)
setup_logger(
    level="DEBUG",
    format="<level>{levelname: <8}</level>{module}:{funcName}:{lineno}:<light_white>{message}</light_white>",
)

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine()
kitten_pic = InputFile.from_path(pathlib.Path("examples/assets/kitten.jpg"))

bot.dispatch.message.auto_rules = IsUser()


async def on_drop(chat_id: int) -> None:
    await api.send_message(chat_id=chat_id, text="Okay, never mind....")


class DummyMiddleware(ABCMiddleware):
    def pre(self) -> bool:
        return True

    def post(self) -> None:
        return None


@bot.on.message(Text("/start"))
async def start(message: Message, me: Me) -> None:
    await message.answer(
        "Hello, {}! It's {}. How are you today?".format(
            message.from_user.first_name,
            me.first_name,
        ),
    )
    m, _ = await wm.wait(
        hasher=MESSAGE_FROM_USER_IN_CHAT(bot.on.message, (message.chat_id, message.from_user.id)),
        release=Text(["fine", "bad"], ignore_case=True),
        lifetime=timedelta(seconds=60),
        on_miss=MessageReplyHandler("Fine or bad", as_reply=True),
        on_drop=partial(on_drop, chat_id=message.chat_id),
    )

    match m.text.unwrap().lower():
        case "fine":
            await m.reply("Cool!")
        case "bad":
            await message.answer_photo(
                photo=kitten_pic,
                caption="I'm sorry... You prob need some kitten pictures",
            )


@bot.on.message(Text("/react"))
async def react(message: Message, context: Context):
    await message.reply("Send me any message...")
    msg, _ = await wm.wait(
        hasher=MESSAGE_FROM_USER_IN_CHAT(bot.on.message, (message.from_user.id, message.chat_id)),
        release=HasText(),
        on_miss=MessageReplyHandler("Your message has no text!"),
        lifespan=DummyMiddleware().to_lifespan(context),
    )
    await msg.react("🔥")


@bot.on.message(Markup("/reverse <text>"))
async def reverse(message: Message, text: str):
    await message.answer(
        text=f"Okay, its.. {text[-1].upper()}.. {text[-2].upper()}.. {text[::-1].lower().capitalize()}",
    )
    if text[::-1].lower().replace(" ", "") == text.lower().replace(" ", ""):
        await message.answer(
            text="Wow.. Seems like this is a palindrome.",
        )


# In order to make argument optional use this trick:
#   1) declare patterns ordered by their broadness
#   2) don't forget to set default value for optional arguments
@bot.on.message(Markup(["/predict", "/predict <thing>"]))
async def predict(thing: str = "it"):
    probability_percent = random.randint(0, 100)
    return f"I predict the probability {thing} will happen is {probability_percent}%"


@bot.on.message(FuzzyText("hello"))
async def hello(message: Message):
    await message.reply("Hi!")


@bot.on.message(FuzzyText("freeze"))
async def freeze_handler(message: Message):
    msg = (
        await message.answer(
            text="Well ok freezing",
            reply_markup=InlineKeyboard().add(InlineButton("Unfreeze", callback_data="unfreeze")).get_markup(),
        )
    ).unwrap()

    with bot.dispatch.middlewares.filter.hold(
        UserId,
        message.from_user.id,
        IsUpdateType(UpdateType.CALLBACK_QUERY),
    ):
        cb, _ = await wm.wait(
            hasher=CALLBACK_QUERY_FOR_MESSAGE(bot.on.callback_query, msg.message_id),
            release=CallbackDataEq("unfreeze"),
        )
        await cb.edit_text("Wow heated!")


bot.run_forever(skip_updates=True)



import typing

from telegrinder.bot.rules.abc import ABCRule, CheckResult

if typing.TYPE_CHECKING:
    from telegrinder.tools.keyboard.button import BaseButton


class ButtonRule[KeyboardButton: BaseButton[typing.Any]](ABCRule):
    def __init__(self, button: KeyboardButton, rule: ABCRule) -> None:
        self.button = button
        self.rule = rule
        self.composable = rule.composable

    @property
    def check(self) -> typing.Callable[..., CheckResult]:
        return self.rule.check


__all__ = ("ButtonRule",)



import typing

from telegrinder.bot.dispatch.view.base import EventModelView
from telegrinder.bot.rules.media import IsMediaGroup
from telegrinder.types.objects import Message

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

    from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager


class MediaGroupView(EventModelView[Message]):
    def __init__(
        self,
        *,
        return_manager: ABCReturnManager | None = None,
        agent_cls: type[Agent] | None = None,
    ) -> None:
        super().__init__(model=Message, return_manager=return_manager, agent_cls=agent_cls)

        self.filter = IsMediaGroup()


__all__ = ("MediaGroupView",)



import typing
from functools import lru_cache

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules.button import ButtonRule
    from telegrinder.tools.keyboard.abc import ABCKeyboard, RawKeyboard
    from telegrinder.tools.keyboard.base import BaseKeyboard
    from telegrinder.tools.keyboard.button import BaseButton


@lru_cache(maxsize=1024)
def get_keyboard_button_rules(
    keyboard_class: type[ABCKeyboard],
    /,
) -> dict[str, ButtonRule[BaseButton[typing.Any]]]:
    from telegrinder.bot.rules.button import ButtonRule

    return {
        name: ButtonRule(button=button, rule=button.rule)
        for name, button in dict(vars(keyboard_class)).items()
        if isinstance(button, keyboard_class.__button_class__) and not is_dunder(name)
    }


def is_dunder(name: str, /) -> bool:
    return name.startswith("__") and name.endswith("__")


def copy_keyboard(keyboard: RawKeyboard, /) -> RawKeyboard:
    return [row.copy() for row in keyboard if row]


def freaky_keyboard_merge[T: BaseKeyboard = typing.Any](
    button: BaseButton[T],
    keyboard_or_button: T | BaseButton[T],
    /,
    *,
    row: bool = False,
) -> T:
    from telegrinder.tools.keyboard.button import BaseButton

    keyboard = button.keyboard_class().add(button)

    if row:
        keyboard = keyboard.row()

    return (
        keyboard.merge_to_last_row(keyboard_or_button)
        if not isinstance(keyboard_or_button, BaseButton)
        else keyboard.add(keyboard_or_button)
    )


def init_keyboard[T: BaseKeyboard = typing.Any](
    keyboard_instance: T,
    /,
    *,
    max_in_row: int = 0,
) -> T:
    for button in get_keyboard_button_rules(type(keyboard_instance)).values():
        if button.button.new_row or (max_in_row and len(keyboard_instance.keyboard[-1]) >= max_in_row):
            keyboard_instance.row()

        keyboard_instance.add(button.button)

    return keyboard_instance


class bound_keyboard_method[T: BaseKeyboard = typing.Any, **P = ..., R = typing.Any]:  # noqa: N801
    def __init__(self, func: typing.Callable[typing.Concatenate[T, P], R], /) -> None:
        self.func = func

    def __get__(self, instance: T | None, owner: type[T]) -> typing.Callable[P, R]:
        return self.func.__get__(instance or owner.__keyboard_instance__, owner)


class RowButtons[KeyboardButton: BaseButton]:
    buttons: typing.Iterable[KeyboardButton]

    def __init__(self, *buttons: KeyboardButton, auto_row: bool = True) -> None:
        self.buttons = buttons
        self.auto_row = auto_row

    def get_data(self) -> list[dict[str, typing.Any]]:
        return [b.get_data() for b in self.buttons]


__all__ = (
    "RowButtons",
    "bound_keyboard_method",
    "copy_keyboard",
    "freaky_keyboard_merge",
    "get_keyboard_button_rules",
    "init_keyboard",
)



import typing
from contextlib import contextmanager

from kungfu.library.monad.result import Error, Ok
from nodnod.agent.base import Agent
from nodnod.agent.event_loop.agent import EventLoopAgent
from nodnod.interface.agent_from_node import create_agent_from_node

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.node.compose import compose

if typing.TYPE_CHECKING:
    from telegrinder.bot.dispatch.process import check_rule
    from telegrinder.bot.rules.abc import ABCRule

else:

    def check_rule(*args: typing.Any, **kwargs: typing.Any) -> bool:
        from telegrinder.bot.dispatch.process import check_rule

        return check_rule(*args, **kwargs)


type AnyNode = typing.Any
type Value = typing.Any


class FilterMiddleware(ABCMiddleware):
    source_filters: dict[AnyNode, tuple[type[Agent], dict[Value, tuple[ABCRule, ...]]]]
    _source_node_agents: dict[AnyNode, Agent]

    def __init__(self) -> None:
        self.source_filters = dict()
        self._source_node_agents = dict()

    async def pre(self, context: Context) -> bool:
        for source_node, (agent_cls, source_filter) in self.source_filters.items():
            if source_node not in self._source_node_agents:
                agent = self._source_node_agents[source_node] = create_agent_from_node(source_node, agent_cls=agent_cls)
            else:
                agent = self._source_node_agents[source_node]

            async with compose(source_node, context, agent=agent) as result:
                match result:
                    case Error(_):
                        return False
                    case Ok(value) if value not in source_filter:
                        return False
                    case _:
                        for filter in source_filter[result.value]:
                            if not await check_rule(filter, context):
                                return False

            return True

        return True

    @contextmanager
    def hold(
        self,
        source_node: AnyNode,
        source_value: Value,
        /,
        *filters: ABCRule,
        agent_cls: type[Agent] | None = None,
    ) -> typing.Generator[None, typing.Any, None]:
        try:
            self.source_filters.setdefault(source_node, (agent_cls or EventLoopAgent, {}))[1][source_value] = filters
            yield
        finally:
            self.release(source_node, source_value)

    def release(self, source_node: AnyNode, source_value: Value, /) -> None:
        source_filter = self.source_filters.get(source_node)

        if source_filter is not None:
            _, source_filter = source_filter
            source_filter.pop(source_value, None)

            if not source_filter:
                self.source_filters.pop(source_node, None)
                self._source_node_agents.pop(source_node, None)
        else:
            self._source_node_agents.pop(source_node, None)


__all__ = ("FilterMiddleware",)



import dataclasses
import operator
import types
import typing

type Operation = GetAttribute | GetItem | MethodCall


def evaluate_operations(
    obj: typing.Any,
    member_name: str,
    operations: list[Operation],
    /,
) -> typing.Any:
    obj = getattr(obj, member_name)

    for operation in operations:
        match operation:
            case GetAttribute(attribute_name=attribute_name):
                obj = getattr(obj, attribute_name)
            case GetItem():
                obj = operation[obj]
            case MethodCall():
                obj = operation(obj)

    return obj


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class MethodCall:
    caller: operator.methodcaller

    def __call__(self, obj: typing.Any, /) -> typing.Any:
        return self.caller(obj)


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class GetAttribute:
    attribute_name: str


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class GetItem:
    itemgetter: operator.itemgetter[typing.Any]

    def __getitem__(self, obj: typing.Any, /) -> typing.Any:
        return self.itemgetter(obj)


class MemberDescriptorProxy:
    _operations: list[Operation]
    _objclass: type[typing.Any]

    __slots__ = ("_descriptor", "_operations", "_member_name", "_objclass")

    def __init__(self, descriptor: types.MemberDescriptorType) -> None:
        self._descriptor = descriptor
        self._member_name = descriptor.__name__
        self._objclass = descriptor.__objclass__
        self._operations = []

    def __getattr__(self, __name: str) -> typing.Any:
        if __name in getattr(type(self), "__static_attributes__"):
            return super().__getattribute__(__name)

        self._operations.append(GetAttribute(attribute_name=__name))
        return self

    def __getitem__(self, __item: typing.Any) -> typing.Any:
        self._operations.append(GetItem(itemgetter=operator.itemgetter(__item)))
        return self

    def __call__(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        if not self._operations or not isinstance(operation := self._operations.pop(-1), GetAttribute):
            raise SyntaxError("No operations to call.")

        self._operations.append(MethodCall(caller=operator.methodcaller(operation.attribute_name, *args, **kwargs)))
        return self

    def __get__(self, instance: typing.Any, owner: typing.Type[typing.Any]) -> typing.Any:
        if instance is None:
            return self
        return self._descriptor.__get__(instance, owner)

    def __set__(self, instance: typing.Any, value: typing.Any) -> None:
        self._descriptor.__set__(instance, value)

    def __delete__(self, instance: typing.Any) -> None:
        self._descriptor.__delete__(instance)

    def __repr__(self) -> str:
        return self._descriptor.__repr__()


__all__ = ("MemberDescriptorProxy", "evaluate_operations")



import typing

from kungfu.library.monad.result import Ok
from nodnod.agent.event_loop.agent import EventLoopAgent

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.node.compose import run_agent
from telegrinder.node.utils import as_node

if typing.TYPE_CHECKING:
    from nodnod.agent.base import Agent

type Node = typing.Any


class NodeRule(ABCRule):
    nodes: tuple[tuple[str | None, Node], ...]

    def __init__(
        self,
        *nodes: Node | tuple[str, Node],
        agent: type[Agent] | None = None,
        roots: dict[type[typing.Any], typing.Any] | None = None,
    ) -> None:
        self.agent = (agent or EventLoopAgent).build(nodes=set(map(as_node, nodes)))
        self.nodes = tuple((x if isinstance(x, tuple) else (None, x)) for x in nodes)
        self.roots = roots

    async def check(self, context: Context) -> bool:
        async with run_agent(self.agent, context, roots=self.roots) as result:
            match result:
                case Ok(scope):
                    for key, node in self.nodes:
                        if key is not None and node in scope:
                            context[key] = scope[node].value

                    return True
                case _:
                    return False


__all__ = ("NodeRule",)



import msgspec

from .objects import ObjectsFieldsLiteralTypesField


class MethodsAnnotationsReturnType(msgspec.Struct):
    name: str
    return_type: str


class MethodsAnnotationsParametersParam(msgspec.Struct):
    name: str
    type: str


class MethodsAnnotationsParameters(msgspec.Struct):
    name: str
    params: list[MethodsAnnotationsParametersParam] = msgspec.field(default_factory=list)


class MethodsAnnotationsLiteralsParam(ObjectsFieldsLiteralTypesField):
    pass


class MethodsAnnotationsLiterals(msgspec.Struct):
    name: str
    params: list[MethodsAnnotationsLiteralsParam] = msgspec.field(default_factory=list)


class MethodsAnnotations(msgspec.Struct):
    return_type: list[MethodsAnnotationsReturnType] = msgspec.field(default_factory=list)
    parameters: list[MethodsAnnotationsParameters] = msgspec.field(default_factory=list)
    literals: list[MethodsAnnotationsLiterals] = msgspec.field(default_factory=list)


class GeneratorMethods(msgspec.Struct):
    annotations: MethodsAnnotations = msgspec.field(default_factory=lambda: MethodsAnnotations())


__all__ = (
    "GeneratorMethods",
    "MethodsAnnotations",
    "MethodsAnnotationsLiterals",
    "MethodsAnnotationsLiteralsParam",
    "MethodsAnnotationsParameters",
    "MethodsAnnotationsParametersParam",
    "MethodsAnnotationsReturnType",
)



import asyncio
import dataclasses
import typing

from kungfu.library.monad.option import Some

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware

type MediaGroupId = str

WAIT_TIME: typing.Final = 2.0
MAX_GROUP_PARTS: typing.Final = 10


@dataclasses.dataclass(frozen=True, slots=True)
class MediaGroupData:
    event: asyncio.Event
    timer: asyncio.TimerHandle
    messages: list[MessageCute] = dataclasses.field(default_factory=list)


class MediaGroupMiddleware(ABCMiddleware):
    media_groups: dict[MediaGroupId, MediaGroupData]

    def __init__(self, *, wait_time: float = WAIT_TIME) -> None:
        self.wait_time = wait_time
        self.media_groups = {}

    async def pre(self, message: MessageCute | None, context: Context) -> bool:
        if message is None:
            return True

        media_group_id = message.media_group_id.unwrap_or_none()

        if not media_group_id:
            return True

        if media_group_id not in self.media_groups:
            media_group_data = MediaGroupData(
                event=(event := asyncio.Event()),
                timer=asyncio.get_running_loop().call_later(self.wait_time, event.set),
                messages=[message],
            )
            self.media_groups[media_group_id] = media_group_data

            await event.wait()

            message.media_group_messages = Some(media_group_data.messages)
            self.media_groups.pop(media_group_id, None)
            return True

        media_group_data = self.media_groups[media_group_id]
        media_group_data.messages.append(message)

        if len(media_group_data.messages) >= MAX_GROUP_PARTS:
            if not media_group_data.timer.cancelled():
                media_group_data.timer.cancel()

            if not media_group_data.event.is_set():
                media_group_data.event.set()

            self.media_groups.pop(media_group_id, None)

        return False


__all__ = ("MediaGroupMiddleware",)



import msgspec


class ObjectsFieldsIdByDefaultField(msgspec.Struct):
    name: str
    nbytes: int


class ObjectsFieldsIdByDefault(msgspec.Struct):
    object_name: str
    fields: list[ObjectsFieldsIdByDefaultField]


class ObjectsFieldsDefaults(msgspec.Struct):
    random_id: list[ObjectsFieldsIdByDefault] = msgspec.field(default_factory=list)


class ObjectsFieldsLiteralTypesEnumLiterals(msgspec.Struct):
    name: str
    enumerations: list[str] = msgspec.field(default_factory=list)


class ObjectsFieldsLiteralTypesField(msgspec.Struct):
    name: str
    literals: list[str | int] = msgspec.field(default_factory=list)
    enum: str | None = None
    enum_literals: ObjectsFieldsLiteralTypesEnumLiterals | None = None
    default: str | None = None

    @property
    def enum_default(self) -> str | None:
        if self.enum is not None and self.default is not None:
            return f"{self.enum}.{self.default}"
        return None

    @property
    def enum_literals_default(self) -> str | None:
        if self.enum_literals is not None and self.default is not None:
            return f"{self.enum_literals.name}.{self.default}"
        return None


class ObjectsFieldsLiteralTypes(msgspec.Struct):
    object_name: str
    fields: list[ObjectsFieldsLiteralTypesField]


class ObjectsFieldsAnnotationsAnnotationsField(msgspec.Struct):
    name: str
    annotation: str
    convert_from: str | None = None


class ObjectsFieldsAnnotationsAnnotations(msgspec.Struct):
    object_name: str
    fields: list[ObjectsFieldsAnnotationsAnnotationsField]


class ObjectsFieldsAnnotations(msgspec.Struct):
    annotations: list[ObjectsFieldsAnnotationsAnnotations] = msgspec.field(default_factory=list)
    literals: list[ObjectsFieldsLiteralTypes] = msgspec.field(default_factory=list)


class ObjectsFields(msgspec.Struct):
    defaults: ObjectsFieldsDefaults = msgspec.field(default_factory=lambda: ObjectsFieldsDefaults())
    annotations: ObjectsFieldsAnnotations = msgspec.field(default_factory=lambda: ObjectsFieldsAnnotations())


class GeneratorObjects(msgspec.Struct):
    fields: ObjectsFields = msgspec.field(default_factory=lambda: ObjectsFields())


__all__ = (
    "GeneratorObjects",
    "ObjectsFields",
    "ObjectsFieldsAnnotations",
    "ObjectsFieldsAnnotationsAnnotations",
    "ObjectsFieldsAnnotationsAnnotationsField",
    "ObjectsFieldsDefaults",
    "ObjectsFieldsIdByDefault",
    "ObjectsFieldsIdByDefaultField",
    "ObjectsFieldsLiteralTypes",
    "ObjectsFieldsLiteralTypesField",
)



"""Example of handling media groups with MediaGroupView.

This example demonstrates how to use the MediaGroupView to handle
media groups sent by users. Media groups are collections
of photos, videos or other media sent together.
"""

from kungfu.library.monad.option import Option

from telegrinder import API, Telegrinder, Token
from telegrinder.modules import setup_logger
from telegrinder.node import Caption, MediaGroup

setup_logger()

api = API(token=Token.from_env())
bot = Telegrinder(api)


@bot.on.media_group()
async def handle_media_group(media_group: MediaGroup, caption: Option[Caption]) -> str:
    return (
        f"📁 Received media group with {len(media_group.items)} items!\n"
        f"🏷️ Media types: {', '.join(set(message.content_type.value for message in media_group.items))}\n"
        f"📝 Caption: {caption.unwrap_or('no caption')}"
    )


bot.run_forever()



import collections
import typing

from telegrinder.tools.formatting.html import FORMATTERS, link, pre_code, tg_emoji
from telegrinder.tools.strings import to_utf16_map, utf16_to_py_index
from telegrinder.types.enums import ContentType, MessageEntityType
from telegrinder.types.methods_utils import get_params
from telegrinder.types.objects import (
    InputFile,
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
    MessageEntity,
    ReactionEmoji,
    ReactionType,
    ReactionTypeEmoji,
)

type InputMedia = typing.Union[
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
]

type MediaType = typing.Literal["animation", "audio", "document", "photo", "video"]

EntitiesDict = collections.defaultdict[int, list[MessageEntity]]

MEDIA_TYPES: typing.Final = (
    ContentType.ANIMATION,
    ContentType.AUDIO,
    ContentType.DOCUMENT,
    ContentType.PHOTO,
    ContentType.VIDEO,
)
INPUT_TYPES: typing.Final = (
    InputMediaAnimation,
    InputMediaAudio,
    InputMediaDocument,
    InputMediaPhoto,
    InputMediaVideo,
)
INPUT_MEDIA_TYPES: typing.Final[dict[ContentType, type[InputMedia]]] = dict(zip(MEDIA_TYPES, INPUT_TYPES))
HTML_FORMAT_ENTITIES: typing.Final[tuple[MessageEntityType, ...]] = (
    MessageEntityType.PRE,
    MessageEntityType.CODE,
    MessageEntityType.TEXT_LINK,
    MessageEntityType.CUSTOM_EMOJI,
    *tuple(x for x in MessageEntityType if x.name.lower() in FORMATTERS),
)


def build_html(text: str, entities: list[MessageEntity], /) -> str:
    utf16_map = to_utf16_map(text)
    text_length_utf16 = utf16_map[-1]
    events = set[int]()

    for entity in entities:
        events.add(entity.offset)
        events.add(entity.offset + entity.length)

    events = sorted(events)
    utf16_to_py = {u: utf16_to_py_index(utf16_map, u) for u in events + [0, text_length_utf16]}

    opens = EntitiesDict(list)
    closes = EntitiesDict(list)

    for entity in entities:
        opens[entity.offset].append(entity)
        closes[entity.offset + entity.length].append(entity)

    result = list[str]()
    stack = collections.deque[MessageEntity]()
    utf16_pos = 0

    while utf16_pos < text_length_utf16:
        py_start = utf16_to_py[utf16_pos]
        next_events = [u for u in events if u > utf16_pos]
        next_utf16 = min(next_events) if next_events else text_length_utf16
        py_end = utf16_to_py[next_utf16]

        if closes[utf16_pos]:
            for _ in closes[utf16_pos]:
                stack.pop()

        if opens[utf16_pos]:
            for e in sorted(opens[utf16_pos], key=lambda e: e.length, reverse=True):
                stack.append(e)

        chunk = text[py_start:py_end]
        formatted = chunk

        for e in reversed(stack):
            if (formatter := FORMATTERS.get(e.type.name.lower())) is not None:
                formatted = formatter(formatted)
            elif e.type in (MessageEntityType.PRE, MessageEntityType.CODE):
                formatted = pre_code(formatted, lang=e.language.unwrap_or_none())
            elif e.type == MessageEntityType.TEXT_LINK:
                formatted = link(e.url.unwrap(), text=formatted)
            elif e.type == MessageEntityType.CUSTOM_EMOJI:
                formatted = tg_emoji(formatted, emoji_id=e.custom_emoji_id.map(int).unwrap())

        result.append(formatted)
        utf16_pos = next_utf16

    return "".join(result)


def compose_reactions(
    reactions: str | ReactionEmoji | ReactionType | list[str | ReactionEmoji | ReactionType],
    /,
) -> list[ReactionType]:
    if not isinstance(reactions, list):
        reactions = [reactions]
    return [
        (
            ReactionTypeEmoji(emoji)
            if isinstance(emoji, ReactionEmoji)
            else (ReactionTypeEmoji(ReactionEmoji(emoji)) if isinstance(emoji, str) else emoji)
        )
        for emoji in reactions
    ]


def input_media(
    type: MediaType,
    media: str | InputFile,
    *,
    caption: str | None = None,
    parse_mode: str | None = None,
    caption_entities: list[MessageEntity] | None = None,
    **other: typing.Any,
) -> InputMedia:
    return INPUT_MEDIA_TYPES[ContentType(type)](**get_params(locals()))


__all__ = ("build_html", "compose_reactions", "input_media")
